#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand
class TensorMaximum(TensorBinOp):
    _op_type_ = OperandDef.MAXIMUM
    _func_name = 'maximum'

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, 'issparse') and x1.issparse() and np.isscalar(x2) and x2 <= 0:
            return True
        if hasattr(x2, 'issparse') and x2.issparse() and np.isscalar(x1) and x1 <= 0:
            return True
        return False


@infer_dtype(np.maximum)
def maximum(x1, x2, out=None, where=None, **kwargs):
    """
    Element-wise maximum of tensor elements.

    Compare two tensors and returns a new array containing the element-wise
    maxima. If one of the elements being compared is a NaN, then that
    element is returned. If both elements are NaNs then the first is
    returned. The latter distinction is important for complex NaNs, which
    are defined as at least one of the real or imaginary parts being a NaN.
    The net effect is that NaNs are propagated.

    Parameters
    ----------
    x1, x2 : array_like
        The tensors holding the elements to be compared. They must have
        the same shape, or shapes that can be broadcast to a single shape.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : ndarray or scalar
        The maximum of `x1` and `x2`, element-wise.  Returns scalar if
        both  `x1` and `x2` are scalars.

    See Also
    --------
    minimum :
        Element-wise minimum of two tensors, propagates NaNs.
    fmax :
        Element-wise maximum of two tensors, ignores NaNs.
    amax :
        The maximum value of a tensor along a given axis, propagates NaNs.
    nanmax :
        The maximum value of a tensor along a given axis, ignores NaNs.

    fmin, amin, nanmin

    Notes
    -----
    The maximum is equivalent to ``mt.where(x1 >= x2, x1, x2)`` when
    neither x1 nor x2 are nans, but it is faster and does proper
    broadcasting.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.maximum([2, 3, 4], [1, 5, 2]).execute()
    array([2, 5, 4])

    >>> mt.maximum(mt.eye(2), [0.5, 2]).execute() # broadcasting
    array([[ 1. ,  2. ],
           [ 0.5,  2. ]])

    >>> mt.maximum([mt.nan, 0, mt.nan], [0, mt.nan, mt.nan]).execute()
    array([ NaN,  NaN,  NaN])
    >>> mt.maximum(mt.Inf, 1).execute()
    inf
    """
    op = TensorMaximum(**kwargs)
    return op(x1, x2, out=out, where=where)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import deque, OrderedDict

from .utils import WorkerActor
from .. import promise
from ..actors import ActorNotExist
from ..utils import mod_hash, log_unhandled

logger = logging.getLogger(__name__)


class DispatchActor(WorkerActor):
    """
    Dispatcher for multiple actors belonging to one category
    """
    def __init__(self):
        super().__init__()
        self._free_slots = dict()
        self._all_slots = dict()
        self._free_slot_requests = dict()
        self._status_ref = None

    def post_create(self):
        super().post_create()

        from .status import StatusActor
        self._status_ref = self.ctx.actor_ref(StatusActor.default_uid())
        if not self.ctx.has_actor(self._status_ref):
            self._status_ref = None

    def _safe_tell_promise_slot(self, callback, queue_name, slot):
        try:
            self.tell_promise(callback, slot)
        except (ActorNotExist, BrokenPipeError, ConnectionRefusedError,
                TimeoutError, promise.PromiseTimeout) as ex:
            from ..scheduler import ResourceActor
            logger.exception('Failed to tell slot %s of queue %s into promise %r',
                             slot, queue_name, callback)

            callback_addr = callback[0][-1]
            if callback_addr != self.address and self.get_schedulers() and \
                    isinstance(ex, (ActorNotExist, BrokenPipeError, ConnectionRefusedError)):
                self.get_actor_ref(ResourceActor.default_uid()).detach_dead_workers([callback_addr], _tell=True)

            if slot is not None:
                self.ref().register_free_slot(slot, queue_name, _tell=True)

    @log_unhandled
    def acquire_free_slot(self, queue_name, callback=None):
        """
        Acquire uid of a free actor when available
        :param queue_name: queue name
        :param callback: promise callback
        """
        if queue_name not in self._free_slots:
            self._safe_tell_promise_slot(callback, queue_name, None)
            return
        if not self._free_slots[queue_name]:
            # no slots free, we queue the callback
            self._free_slot_requests[queue_name].append(callback)
            logger.debug('Slots not enough for queue %s on %s. slot dump: %r',
                         queue_name, self.address, self.get_free_slots_num())
            return

        free_slot = self._free_slots[queue_name].popitem()[0]
        logger.debug('Slot %s allocated for queue %s on %s. slot dump: %r',
                     free_slot, queue_name, self.address, self.get_free_slots_num())
        self._safe_tell_promise_slot(callback, queue_name, free_slot)

        if self._status_ref is not None:
            self._status_ref.update_slots({queue_name: len(self._free_slots[queue_name])},
                                          _tell=True, _wait=False)

    @log_unhandled
    def get_hash_slot(self, queue_name, key):
        """
        Get uid of a slot by hash value
        :param queue_name: queue name
        :param key: key to be hashed
        """
        slots = list(self._all_slots[queue_name])
        uid = slots[mod_hash(key, len(slots))]
        return uid

    @log_unhandled
    def get_hash_slots(self, queue_name, keys):
        slots = list(self._all_slots[queue_name])
        uids = [slots[mod_hash(key, len(slots))] for key in keys]
        return uids

    @log_unhandled
    def get_slots(self, queue_name):
        """
        Get all uids of slots of a queue
        :param queue_name: queue name
        """
        if queue_name not in self._all_slots:
            return []
        return list(self._all_slots[queue_name].keys())

    def get_free_slots_num(self):
        """
        Get number of free slots of every queue
        """
        return dict((k, len(v)) for k, v in self._free_slots.items())

    @log_unhandled
    def register_free_slot(self, uid, queue_name):
        """
        Register free uid of a queue
        :param uid: uid of free actor
        :param queue_name: queue name
        """
        if queue_name not in self._free_slots:
            self._free_slot_requests[queue_name] = deque()
            self._free_slots[queue_name] = OrderedDict()
            self._all_slots[queue_name] = OrderedDict()
        self._free_slots[queue_name][uid] = ''
        self._all_slots[queue_name][uid] = ''

        logger.debug('Slot %s registered for queue %s on %s', uid, queue_name, self.address)

        if self._free_slot_requests[queue_name]:
            free_slot = self._free_slots[queue_name].popitem()[0]
            logger.debug('Slot %s allocated for queue %s on %s. slot dump: %r',
                         free_slot, queue_name, self.address, self.get_free_slots_num())
            self._safe_tell_promise_slot(self._free_slot_requests[queue_name].popleft(), queue_name, free_slot)

        if self._status_ref is not None:
            self._status_ref.update_slots({queue_name: len(self._free_slots[queue_name])},
                                          _tell=True, _wait=False)

    @log_unhandled
    def remove_slot(self, uid, queue_name):
        self._free_slots[queue_name].pop(uid, None)
        self._all_slots[queue_name].pop(uid, None)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...serialize import KeyField, Float64Field, Int8Field
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape, recursive_tile
from ..array_utils import as_same_device, device
from ..core import TensorOrder
from ..datasource import tensor as astensor
from ..operands import TensorOperand, TensorOperandMixin
from ..utils import validate_axis


class TensorTrapz(TensorOperand, TensorOperandMixin):
    _op_type_ = opcodes.TRAPZ

    _y = KeyField('y')
    _x = KeyField('x')
    _dx = Float64Field('dx')
    _axis = Int8Field('axis')

    def __init__(self, y=None, x=None, dx=None, axis=None, **kw):
        super().__init__(_y=y, _x=x, _dx=dx, _axis=axis, **kw)

    @property
    def y(self):
        return self._y

    @property
    def x(self):
        return self._x

    @property
    def dx(self):
        return self._dx

    @property
    def axis(self):
        return self._axis

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._y = self._inputs[0]
        if self._x is not None:
            self._x = self._inputs[-1]

    def __call__(self, y, x=None):
        inputs = [y]
        order = y.order
        if x is not None:
            x = astensor(x)
            inputs.append(x)
            if x.order == TensorOrder.C_ORDER:
                order = TensorOrder.C_ORDER

        shape = tuple(s for ax, s in enumerate(y.shape)
                      if ax != self._axis)
        dtype = np.trapz(np.empty(1, dtype=y.dtype)).dtype
        return self.new_tensor(inputs, shape=shape, dtype=dtype,
                               order=order)

    @classmethod
    def tile(cls, op: "TensorTrapz"):
        from .diff import diff

        y = astensor(op.y)
        x = op.x
        axis = op.axis

        if x is not None:
            x = astensor(x)
            # rechunk x to make x.nsplits == y.nsplits
            check_chunks_unknown_shape([x, y], TilesError)
            x = x.rechunk(y.nsplits)._inplace_tile()

        if len(y.chunks) == 1:
            return cls._tile_one_chunk(op, y, x)

        if x is None:
            d = op.dx
        else:
            if x.ndim == 1:
                d = diff(x)
                # reshape to correct shape
                shape = [1]*y.ndim
                shape[axis] = d.shape[0]
                d = d.reshape(shape)
            else:
                d = diff(x, axis=axis)
        nd = y.ndim
        slice1 = [slice(None)]*nd
        slice2 = [slice(None)]*nd
        slice1[axis] = slice(1, None)
        slice2[axis] = slice(None, -1)
        ret = (d * (y[tuple(slice1)] + y[tuple(slice2)]) / 2.0).sum(axis)
        return [recursive_tile(ret)]

    @classmethod
    def _tile_one_chunk(cls, op, y, x):
        out = op.outputs[0]
        chunk_op = op.copy().reset_key()
        inputs = [y.chunks[0]]
        if x is not None:
            inputs.append(x.chunks[0])
        chunk = chunk_op.new_chunk(inputs, shape=out.shape,
                                   order=out.order,
                                   index=(0,) * out.ndim)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out.shape, order=out.order,
                                  nsplits=tuple((s,) for s in out.shape),
                                  chunks=[chunk])

    @classmethod
    def execute(cls, ctx, op: "TensorTrapz"):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)

        y = inputs[0]
        if len(inputs) > 1:
            x = inputs[-1]
        else:
            x = None

        with device(device_id):
            ctx[op.outputs[0].key] = xp.trapz(y, x=x, dx=op.dx,
                                              axis=op.axis)


def trapz(y, x=None, dx=1.0, axis=-1):
    """
    Integrate along the given axis using the composite trapezoidal rule.

    Integrate `y` (`x`) along given axis.

    Parameters
    ----------
    y : array_like
        Input tensor to integrate.
    x : array_like, optional
        The sample points corresponding to the `y` values. If `x` is None,
        the sample points are assumed to be evenly spaced `dx` apart. The
        default is None.
    dx : scalar, optional
        The spacing between sample points when `x` is None. The default is 1.
    axis : int, optional
        The axis along which to integrate.

    Returns
    -------
    trapz : float
        Definite integral as approximated by trapezoidal rule.

    See Also
    --------
    sum, cumsum

    Notes
    -----
    Image [2]_ illustrates trapezoidal rule -- y-axis locations of points
    will be taken from `y` tensor, by default x-axis distances between
    points will be 1.0, alternatively they can be provided with `x` tensor
    or with `dx` scalar.  Return value will be equal to combined area under
    the red lines.


    References
    ----------
    .. [1] Wikipedia page: https://en.wikipedia.org/wiki/Trapezoidal_rule

    .. [2] Illustration image:
           https://en.wikipedia.org/wiki/File:Composite_trapezoidal_rule_illustration.png

    Examples
    --------
    >>> import mars.tensor as mt
    >>> mt.trapz([1,2,3]).execute()
    4.0
    >>> mt.trapz([1,2,3], x=[4,6,8]).execute()
    8.0
    >>> mt.trapz([1,2,3], dx=2).execute()
    8.0
    >>> a = mt.arange(6).reshape(2, 3)
    >>> a.execute()
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> mt.trapz(a, axis=0).execute()
    array([1.5, 2.5, 3.5])
    >>> mt.trapz(a, axis=1).execute()
    array([2.,  8.])

    """
    y = astensor(y)
    axis = validate_axis(y.ndim, axis)
    op = TensorTrapz(y=y, x=x, dx=dx, axis=axis)
    return op(y, x=x)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...custom_log import redirect_custom_log
from ...serialize import TupleField, DictField, FunctionField, StringField
from ...utils import enter_current_session, quiet_stdio
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import build_empty_df, build_empty_series, parse_index, \
    validate_output_types, make_dtypes, make_dtype


class GroupByApply(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.APPLY
    _op_module_ = 'dataframe.groupby'

    _func = FunctionField('func')
    _args = TupleField('args')
    _kwds = DictField('kwds')
    # for chunk
    _tileable_op_key = StringField('tileable_op_key')

    def __init__(self, func=None, args=None, kwds=None, output_types=None,
                 tileable_op_key=None, **kw):
        super().__init__(_func=func, _args=args, _kwds=kwds, _output_types=output_types,
                         _tileable_op_key=tileable_op_key, **kw)

    @property
    def func(self):
        return self._func

    @property
    def args(self):
        return getattr(self, '_args', None) or ()

    @property
    def kwds(self):
        return getattr(self, '_kwds', None) or dict()

    @property
    def tileable_op_key(self):
        return self._tileable_op_key

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op):
        in_data = ctx[op.inputs[0].key]
        out = op.outputs[0]
        if not in_data:
            if op.output_types[0] == OutputType.dataframe:
                ctx[op.outputs[0].key] = build_empty_df(op.outputs[0].dtypes)
            else:
                ctx[op.outputs[0].key] = build_empty_series(
                    op.outputs[0].dtype, name=out.name)
            return

        applied = in_data.apply(op.func, *op.args, **op.kwds)

        if isinstance(applied, pd.DataFrame):
            # when there is only one group, pandas tend to return a DataFrame, while
            # we need to convert it into a compatible series
            if op.output_types[0] == OutputType.series:
                assert len(applied.index) == 1
                applied_idx = pd.MultiIndex.from_arrays(
                    [[applied.index[0]] * len(applied.columns), applied.columns.tolist()])
                applied_idx.names = [applied.index.name, None]
                applied = pd.Series(np.array(applied.iloc[0]), applied_idx,
                                    name=applied.columns.name)
            else:
                applied.columns.name = None
        else:
            applied.name = out.name
        ctx[out.key] = applied

    @classmethod
    def tile(cls, op):
        in_groupby = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_groupby.chunks:
            inp_chunks = [c]

            new_op = op.copy().reset_key()
            new_op._tileable_op_key = op.key
            if op.output_types[0] == OutputType.dataframe:
                chunks.append(new_op.new_chunk(
                    inp_chunks, index=c.index, shape=(np.nan, len(out_df.dtypes)), dtypes=out_df.dtypes,
                    columns_value=out_df.columns_value, index_value=out_df.index_value))
            else:
                chunks.append(new_op.new_chunk(
                    inp_chunks, name=out_df.name, index=(c.index[0],), shape=(np.nan,), dtype=out_df.dtype,
                    index_value=out_df.index_value))

        new_op = op.copy()
        kw = out_df.params.copy()
        kw['chunks'] = chunks
        if op.output_types[0] == OutputType.dataframe:
            kw['nsplits'] = ((np.nan,) * len(chunks), (out_df.shape[1],))
        else:
            kw['nsplits'] = ((np.nan,) * len(chunks),)
        return new_op.new_tileables([in_groupby], **kw)

    def _infer_df_func_returns(self, in_groupby, in_df, dtypes, dtype=None,
                               name=None, index=None):
        index_value, output_type, new_dtypes = None, None, None

        try:
            infer_df = in_groupby.op.build_mock_groupby().apply(self.func, *self.args, **self.kwds)

            # todo return proper index when sort=True is implemented
            index_value = parse_index(infer_df.index[:0], in_df.key, self.func)

            # for backward compatibility
            dtype = dtype if dtype is not None else dtypes
            if isinstance(infer_df, pd.DataFrame):
                output_type = output_type or OutputType.dataframe
                new_dtypes = new_dtypes or infer_df.dtypes
            elif isinstance(infer_df, pd.Series):
                output_type = output_type or OutputType.series
                new_dtypes = new_dtypes or (name or infer_df.name, dtype or infer_df.dtype)
            else:
                output_type = OutputType.series
                new_dtypes = (name, dtype or pd.Series(infer_df).dtype)
        except:  # noqa: E722  # nosec
            pass

        self.output_types = [output_type] if not self.output_types else self.output_types
        dtypes = new_dtypes if dtypes is None else dtypes
        index_value = index_value if index is None else parse_index(index)
        return dtypes, index_value

    def __call__(self, groupby, dtypes=None, dtype=None, name=None, index=None):
        in_df = groupby
        while in_df.op.output_types[0] not in (OutputType.dataframe, OutputType.series):
            in_df = in_df.inputs[0]

        with quiet_stdio():
            dtypes, index_value = self._infer_df_func_returns(
                groupby, in_df, dtypes, dtype=dtype, name=name, index=index)
        if index_value is None:
            index_value = parse_index(None, (in_df.key, in_df.index_value.key))
        for arg, desc in zip((self.output_types, dtypes), ('output_types', 'dtypes')):
            if arg is None:
                raise TypeError(f'Cannot determine {desc} by calculating with enumerate data, '
                                'please specify it as arguments')

        if self.output_types[0] == OutputType.dataframe:
            new_shape = (np.nan, len(dtypes))
            return self.new_dataframe([groupby], shape=new_shape, dtypes=dtypes, index_value=index_value,
                                      columns_value=parse_index(dtypes.index, store_data=True))
        else:
            name = name or dtypes[0]
            dtype = dtype or dtypes[1]
            new_shape = (np.nan,)
            return self.new_series([groupby], name=name, shape=new_shape, dtype=dtype,
                                   index_value=index_value)


def groupby_apply(groupby, func, *args, output_type=None, dtypes=None, dtype=None,
                  name=None, index=None, **kwargs):
    """
    Apply function `func` group-wise and combine the results together.

    The function passed to `apply` must take a dataframe as its first
    argument and return a DataFrame, Series or scalar. `apply` will
    then take care of combining the results back together into a single
    dataframe or series. `apply` is therefore a highly flexible
    grouping method.

    While `apply` is a very flexible method, its downside is that
    using it can be quite a bit slower than using more specific methods
    like `agg` or `transform`. Pandas offers a wide range of method that will
    be much faster than using `apply` for their specific purposes, so try to
    use them before reaching for `apply`.

    Parameters
    ----------
    func : callable
        A callable that takes a dataframe as its first argument, and
        returns a dataframe, a series or a scalar. In addition the
        callable may take positional and keyword arguments.

    output_type : {'dataframe', 'series'}, default None
        Specify type of returned object. See `Notes` for more details.

    dtypes : Series, default None
        Specify dtypes of returned DataFrames. See `Notes` for more details.

    dtype : numpy.dtype, default None
        Specify dtype of returned Series. See `Notes` for more details.

    name : str, default None
        Specify name of returned Series. See `Notes` for more details.

    index : Index, default None
        Specify index of returned object. See `Notes` for more details.

    args, kwargs : tuple and dict
        Optional positional and keyword arguments to pass to `func`.

    Returns
    -------
    applied : Series or DataFrame

    See Also
    --------
    pipe : Apply function to the full GroupBy object instead of to each
        group.
    aggregate : Apply aggregate function to the GroupBy object.
    transform : Apply function column-by-column to the GroupBy object.
    Series.apply : Apply a function to a Series.
    DataFrame.apply : Apply a function to each row or column of a DataFrame.

    Notes
    -----
    When deciding output dtypes and shape of the return value, Mars will
    try applying ``func`` onto a mock grouped object, and the apply call
    may fail. When this happens, you need to specify the type of apply
    call (DataFrame or Series) in output_type.

    * For DataFrame output, you need to specify a list or a pandas Series
      as ``dtypes`` of output DataFrame. ``index`` of output can also be
      specified.
    * For Series output, you need to specify ``dtype`` and ``name`` of
      output Series.
    """
    output_types = kwargs.pop('output_types', None)
    object_type = kwargs.pop('object_type', None)
    output_types = validate_output_types(
        output_types=output_types, output_type=output_type, object_type=object_type)

    dtypes = make_dtypes(dtypes)
    dtype = make_dtype(dtype)
    op = GroupByApply(func=func, args=args, kwds=kwargs, output_types=output_types)
    return op(groupby, dtypes=dtypes, dtype=dtype, name=name, index=index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import ValueType, KeyField, ListField
from ..operands import TensorHasInput, TensorOperandMixin
from ..datasource import tensor as astensor
from ..array_utils import as_same_device, device
from ..utils import reverse_order
from ..core import TensorOrder


def _reorder(x, axes):
    if x is None:
        return
    return type(x)(np.array(x)[list(axes)].tolist())


class TensorTranspose(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.TRANSPOSE

    _input = KeyField('input')
    _axes = ListField('axes', ValueType.int32)

    def __init__(self, axes=None, dtype=None, sparse=False, **kw):
        super().__init__(_axes=axes, _dtype=dtype, _sparse=sparse,
                         # transpose will create a view
                         _create_view=True, **kw)

    @property
    def axes(self):
        return getattr(self, '_axes', None)

    def __call__(self, a):
        shape = tuple(s if np.isnan(s) else int(s)
                      for s in _reorder(a.shape, self._axes))
        if self._axes == list(reversed(range(a.ndim))):
            # order reversed
            tensor_order = reverse_order(a.order)
        else:
            tensor_order = TensorOrder.C_ORDER
        return self.new_tensor([a], shape, order=tensor_order)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def on_output_modify(self, new_output):
        op = self.copy().reset_key()
        return op(new_output)

    def on_input_modify(self, new_input):
        op = self.copy().reset_key()
        return op(new_input)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        out_chunks = []
        for c in op.inputs[0].chunks:
            chunk_op = op.copy().reset_key()
            chunk_shape = tuple(s if np.isnan(s) else int(s)
                                for s in _reorder(c.shape, op.axes))
            chunk_idx = _reorder(c.index, op.axes)
            out_chunk = chunk_op.new_chunk([c], shape=chunk_shape,
                                           index=chunk_idx, order=tensor.order)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        nsplits = _reorder(op.inputs[0].nsplits, op.axes)
        return new_op.new_tensors(op.inputs, op.outputs[0].shape, order=tensor.order,
                                  chunks=out_chunks, nsplits=nsplits)

    @classmethod
    def execute(cls, ctx, op):
        (x,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)

        axes = op.axes
        with device(device_id):
            ctx[op.outputs[0].key] = xp.transpose(x, axes or None)


def transpose(a, axes=None):
    """
    Permute the dimensions of a tensor.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axes : list of ints, optional
        By default, reverse the dimensions, otherwise permute the axes
        according to the values given.

    Returns
    -------
    p : Tensor
        `a` with its axes permuted.  A view is returned whenever
        possible.

    See Also
    --------
    moveaxis
    argsort

    Notes
    -----
    Use `transpose(a, argsort(axes))` to invert the transposition of tensors
    when using the `axes` keyword argument.

    Transposing a 1-D array returns an unchanged view of the original tensor.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(4).reshape((2,2))
    >>> x.execute()
    array([[0, 1],
           [2, 3]])

    >>> mt.transpose(x).execute()
    array([[0, 2],
           [1, 3]])

    >>> x = mt.ones((1, 2, 3))
    >>> mt.transpose(x, (1, 0, 2)).shape
    (2, 1, 3)

    """
    a = astensor(a)
    if axes:
        if len(axes) != a.ndim:
            raise ValueError("axes don't match tensor")

    if not axes:
        axes = list(range(a.ndim))[::-1]
    else:
        axes = list(axes)
    op = TensorTranspose(axes, dtype=a.dtype, sparse=a.issparse())
    return op(a)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    import pyproxima2 as proxima
except ImportError:  # pragma: no cover
    proxima = None

from ... import tensor as mt
from ...tensor.merge import TensorConcatenate
from ...tensor.indexing import TensorSlice


available_numpy_dtypes = [
    np.dtype(np.float16),
    np.dtype(np.float32),
    np.dtype(np.int8),
    np.dtype(np.int16),
]


if proxima:
    _proxima_types = [
        proxima.IndexMeta.FT_FP16,
        proxima.IndexMeta.FT_FP32,
        proxima.IndexMeta.FT_INT8,
        proxima.IndexMeta.FT_INT16,
    ]
    assert len(_proxima_types) == len(available_numpy_dtypes)
    _type_mapping = {numpy_dtype: proxima_type
                     for numpy_dtype, proxima_type
                     in zip(available_numpy_dtypes, _proxima_types)}


def rechunk_tensor(tensor, chunk_size):
    # TODO(hks): Provide a unify rechunk logic with mmap.
    cur_chunks = []

    out_nchunks = tensor.shape[0] // chunk_size
    row_nsplits = [chunk_size] * out_nchunks
    rest = tensor.shape[0] % chunk_size
    if rest >= out_nchunks:
        row_nsplits.append(rest)
    else:
        for i in range(tensor.shape[0] % chunk_size):
            row_nsplits[-i-1] += 1

    tensor_cumnrows = np.cumsum([0] + list(tensor.nsplits[0]))
    offset = 0
    out_groups = []
    for split in row_nsplits:
        start_chunk_index = int(tensor_cumnrows.searchsorted(offset))
        start_chunk_index = start_chunk_index - 1 if start_chunk_index != 0 else 0
        end_chunk_index = int(tensor_cumnrows.searchsorted(offset + split) - 1)
        if start_chunk_index == end_chunk_index:
            t = tensor.chunks[start_chunk_index]
            slice_op = TensorSlice((slice(offset - tensor_cumnrows[start_chunk_index],
                                          split + offset - tensor_cumnrows[end_chunk_index]),
                                    slice(None)), dtype=t.dtype)
            out_groups.append([slice_op.new_chunk([t], shape=(split, t.shape[1]),
                                                  index=(len(cur_chunks), 0),
                                                  order=t.order)])
        else:
            chunks = []
            start_chunk = tensor.chunks[start_chunk_index]
            start_slice = int(offset - tensor_cumnrows[start_chunk_index])
            slice_op = TensorSlice((slice(start_slice, None),
                                    slice(None)), dtype=start_chunk.dtype)
            chunks.append(slice_op.new_chunk([start_chunk], shape=(start_chunk.shape[0] - start_slice,
                                                                   start_chunk.shape[1]),
                                             index=(0, 0),
                                             order=start_chunk.order))
            chunks.extend(tensor.chunks[start_chunk_index + 1: end_chunk_index])
            end_chunk = tensor.chunks[end_chunk_index]
            end_slice = int(split + offset - tensor_cumnrows[end_chunk_index])
            slice_op_end = TensorSlice((slice(None, end_slice),
                                        slice(None)), dtype=start_chunk.dtype)
            chunks.append(slice_op_end.new_chunk([end_chunk], shape=(end_slice, end_chunk.shape[1]),
                                                 index=(end_chunk_index - start_chunk_index, 0),
                                                 order=end_chunk.order))
            out_groups.append(chunks)

        offset += split

    return out_groups


def build_mmap_chunks(chunks, worker, file_prefix):
    write_mmap_chunks = []
    nrows = sum(c.shape[0] for c in chunks)
    array_shape = (nrows, chunks[0].shape[1])
    array_dtype = chunks[0].dtype
    create_mmap_op = TensorConcatenate(mmap=True, create_mmap_file=True,
                                       total_shape=array_shape,
                                       file_prefix=file_prefix,
                                       dtype=array_dtype)
    create_mmap_op._expect_worker = worker
    create_mmap_chunk = create_mmap_op.new_chunk(
        None, index=(0,), shape=(), dtype=array_dtype)
    start_index = 0
    for j, chk in enumerate(chunks):
        s = slice(start_index, start_index + chk.shape[0])
        start_index += chk.shape[0]
        write_mmap_op = TensorConcatenate(mmap=True, create_mmap_file=False,
                                          total_shape=array_shape,
                                          partition_slice=s,
                                          dtype=array_dtype)
        write_mmap_op._expect_worker = worker
        write_mmap_chunk = write_mmap_op.new_chunk([create_mmap_chunk, chk],
                                                   index=(j + 1, 0), shape=(),
                                                   dtype=array_dtype)
        write_mmap_chunks.append(write_mmap_chunk)
    return write_mmap_chunks


def validate_tensor(tensor):
    if hasattr(tensor, 'to_tensor'):
        tensor = tensor.to_tensor()
    else:
        tensor = mt.tensor(tensor)
    if tensor.ndim != 2:
        raise ValueError('Input tensor should be 2-d')
    return tensor


def get_proxima_type(np_dtype):
    try:
        return _type_mapping[np_dtype]
    except KeyError:
        raise TypeError(f"Does not support {np_dtype}, available types include "
                        f"{', '.join(t.name for t in _type_mapping)}")



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utils import infer_dtype, implement_scipy
from .core import spspecial, TensorSpecialBinOp


class TensorXLogY(TensorSpecialBinOp):
    _func_name = 'xlogy'

    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, 'issparse') and x1.issparse():
            return True
        return False


@implement_scipy(spspecial.xlogy)
@infer_dtype(spspecial.xlogy)
def xlogy(x1, x2, out=None, where=None, **kwargs):
    op = TensorXLogY(**kwargs)
    return op(x1, x2, out=out, where=where)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bisect
import itertools
import time
import uuid
from collections import defaultdict, deque
from enum import Enum

from ..config import options
from ..utils import tokenize
from .utils import WorkerActor


class EventCategory(Enum):
    RESOURCE = 0
    PROCEDURE = 1


class EventLevel(Enum):
    NORMAL = 0
    WARNING = 1
    ERROR = 2
    CRITICAL = 3


class ResourceEventType(Enum):
    MEM_HIGH = 0


class ProcedureEventType(Enum):
    __order__ = 'CPU_CALC GPU_CALC DISK_IO NETWORK'

    CPU_CALC = 0
    GPU_CALC = 1
    DISK_IO = 2
    NETWORK = 3


class WorkerEvent(object):
    __slots__ = 'event_id', 'category', 'level', 'event_type', 'owner', 'time_start', 'time_end'

    def __init__(self, category=None, level=None, event_type=None, owner=None,
                 time_start=None, time_end=None, event_id=None):
        self.category = category
        self.level = level or EventLevel.NORMAL
        self.event_type = event_type
        self.owner = owner
        self.time_start = time_start
        self.time_end = time_end

        self.event_id = event_id or tokenize(
            uuid.getnode(), time.time(), category, level, event_type, owner)


class EventsActor(WorkerActor):
    def __init__(self):
        super().__init__()
        self._event_timelines = defaultdict(deque)
        self._id_to_open_event = dict()

    def add_single_event(self, category, level, event_type, owner=None):
        event_obj = WorkerEvent(
            category=category, level=level, event_type=event_type, owner=owner,
            time_start=time.time(), time_end=time.time())
        self._event_timelines[category].append((time.time(), event_obj))

        self._purge_old_events(category)
        return event_obj.event_id

    def add_open_event(self, category, level, event_type, owner=None):
        event_obj = WorkerEvent(
            category=category, level=level, event_type=event_type, owner=owner,
            time_start=time.time(), time_end=None)
        self._event_timelines[category].append((time.time(), event_obj))
        self._id_to_open_event[event_obj.event_id] = event_obj

        self._purge_old_events(category)
        return event_obj.event_id

    def close_event(self, event_id):
        try:
            event_obj = self._id_to_open_event.pop(event_id)
        except KeyError:
            return
        event_obj.time_end = time.time()
        self._event_timelines[event_obj.category].append((time.time(), event_obj))
        self._purge_old_events(event_obj.category)

    def query_by_time(self, category, time_start=None, time_end=None):
        self._purge_old_events(category)

        class ItemWrapper(object):
            def __init__(self, tl):
                self._l = tl

            def __getitem__(self, item):
                return self._l[item][0]

            def __len__(self):
                return len(self._l)

        timeline = self._event_timelines[category]
        left_pos = 0 if time_start is None \
            else bisect.bisect_left(ItemWrapper(timeline), time_start)
        right_pos = len(timeline) if time_end is None \
            else bisect.bisect_right(ItemWrapper(timeline), time_end)
        return [it[1] for it in itertools.islice(timeline, left_pos, right_pos)]

    def _purge_old_events(self, category):
        check_time = time.time()
        min_accept_time = check_time - options.worker.event_preserve_time

        timeline = self._event_timelines[category]
        while timeline and timeline[0][1].time_end is not None \
                and timeline[0][1].time_end < min_accept_time:
            timeline.popleft()


class EventContext(object):
    def __init__(self, events_ref, category, level, event_type, owner=None):
        self._events_ref = events_ref
        if events_ref is not None:
            self._event_id = events_ref.add_open_event(category, level, event_type, owner)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._events_ref is not None:
            self._events_ref.close_event(self._event_id)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialize import KeyField, StringField, TupleField, DictField, BoolField
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_empty_series


class SeriesDatetimeMethod(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATETIME_METHOD

    _input = KeyField('input')
    _method = StringField('method')
    _method_args = TupleField('method_args')
    _method_kwargs = DictField('method_kwargs')
    _is_property = BoolField('is_property')

    def __init__(self, method=None, method_args=None, method_kwargs=None,
                 is_property=None, stage=None, output_types=None, **kw):
        super().__init__(_method=method, _method_args=method_args,
                         _method_kwargs=method_kwargs, _is_property=is_property,
                         _stage=stage, _output_types=output_types, **kw)
        if not self.output_types:
            self.output_types = [OutputType.series]

    @property
    def input(self):
        return self._input

    @property
    def method(self):
        return self._method

    @property
    def method_args(self):
        return self._method_args

    @property
    def method_kwargs(self):
        return self._method_kwargs

    @property
    def is_property(self):
        return self._is_property

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, inp):
        return _datetime_method_to_handlers[self._method].call(self, inp)

    @classmethod
    def tile(cls, op):
        return _datetime_method_to_handlers[op.method].tile(op)

    @classmethod
    def execute(cls, ctx, op):
        return _datetime_method_to_handlers[op.method].execute(ctx, op)


class SeriesDatetimeMethodBaseHandler:
    @classmethod
    def call(cls, op, inp):
        empty_series = build_empty_series(inp.dtype)
        if op.is_property:
            test_obj = getattr(empty_series.dt, op.method)
        else:
            test_obj = getattr(empty_series.dt, op.method)(
                *op.method_args, **op.method_kwargs)
        dtype = test_obj.dtype
        return op.new_series([inp], shape=inp.shape,
                             dtype=dtype, index_value=inp.index_value,
                             name=inp.name)

    @classmethod
    def tile(cls, op):
        out = op.outputs[0]

        out_chunks = []
        for series_chunk in op.input.chunks:
            chunk_op = op.copy().reset_key()
            out_chunks.append(chunk_op.new_chunk(
                [series_chunk], shape=series_chunk.shape,
                dtype=out.dtype, index=series_chunk.index,
                index_value=series_chunk.index_value,
                name=series_chunk.name))

        params = out.params
        params['chunks'] = out_chunks
        params['nsplits'] = op.input.nsplits
        new_op = op.copy()
        return new_op.new_tileables([op.input], kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        inp = ctx[op.input.key]
        try:
            out = getattr(inp.dt, op.method)
        except ValueError:
            # fail due to buffer read-only
            out = getattr(inp.copy().dt, op.method)
        if not op.is_property:
            out = out(*op.method_args, **op.method_kwargs)
        ctx[op.outputs[0].key] = out


_datetime_method_to_handlers = {}
for method in dir(pd.Series.dt):
    if not method.startswith('_'):
        _datetime_method_to_handlers[method] = SeriesDatetimeMethodBaseHandler



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

import numpy as np

try:
    from scipy.stats import distributions as sp_distributions
except ImportError:
    sp_distributions = None

from ...core import ExecutableTuple
from ...utils import require_not_none
from .. import special
from ..datasource import asarray


# Map from names to lambda_ values used in power_divergence().
_power_div_lambda_names = {
    "pearson": 1,
    "log-likelihood": 0,
    "freeman-tukey": -0.5,
    "mod-log-likelihood": -1,
    "neyman": -2,
    "cressie-read": 2/3,
}


def _count(a, axis=None):
    if axis is None:
        return a.size
    else:
        return a.shape[axis]


Power_divergenceResult = namedtuple('Power_divergenceResult',
                                    ('statistic', 'pvalue'))


@require_not_none(sp_distributions)
def power_divergence(f_obs, f_exp=None, ddof=0, axis=0, lambda_=None):
    """
    Cressie-Read power divergence statistic and goodness of fit test.

    This function tests the null hypothesis that the categorical data
    has the given frequencies, using the Cressie-Read power divergence
    statistic.

    Parameters
    ----------
    f_obs : array_like
        Observed frequencies in each category.
    f_exp : array_like, optional
        Expected frequencies in each category.  By default the categories are
        assumed to be equally likely.
    ddof : int, optional
        "Delta degrees of freedom": adjustment to the degrees of freedom
        for the p-value.  The p-value is computed using a chi-squared
        distribution with ``k - 1 - ddof`` degrees of freedom, where `k`
        is the number of observed frequencies.  The default value of `ddof`
        is 0.
    axis : int or None, optional
        The axis of the broadcast result of `f_obs` and `f_exp` along which to
        apply the test.  If axis is None, all values in `f_obs` are treated
        as a single data set.  Default is 0.
    lambda_ : float or str, optional
        The power in the Cressie-Read power divergence statistic.  The default
        is 1.  For convenience, `lambda_` may be assigned one of the following
        strings, in which case the corresponding numerical value is used::

            String              Value   Description
            "pearson"             1     Pearson's chi-squared statistic.
                                        In this case, the function is
                                        equivalent to `stats.chisquare`.
            "log-likelihood"      0     Log-likelihood ratio. Also known as
                                        the G-test [3]_.
            "freeman-tukey"      -1/2   Freeman-Tukey statistic.
            "mod-log-likelihood" -1     Modified log-likelihood ratio.
            "neyman"             -2     Neyman's statistic.
            "cressie-read"        2/3   The power recommended in [5]_.

    Returns
    -------
    statistic : float or ndarray
        The Cressie-Read power divergence test statistic.  The value is
        a float if `axis` is None or if` `f_obs` and `f_exp` are 1-D.
    pvalue : float or ndarray
        The p-value of the test.  The value is a float if `ddof` and the
        return value `stat` are scalars.

    See Also
    --------
    chisquare

    Notes
    -----
    This test is invalid when the observed or expected frequencies in each
    category are too small.  A typical rule is that all of the observed
    and expected frequencies should be at least 5.

    When `lambda_` is less than zero, the formula for the statistic involves
    dividing by `f_obs`, so a warning or error may be generated if any value
    in `f_obs` is 0.

    Similarly, a warning or error may be generated if any value in `f_exp` is
    zero when `lambda_` >= 0.

    The default degrees of freedom, k-1, are for the case when no parameters
    of the distribution are estimated. If p parameters are estimated by
    efficient maximum likelihood then the correct degrees of freedom are
    k-1-p. If the parameters are estimated in a different way, then the
    dof can be between k-1-p and k-1. However, it is also possible that
    the asymptotic distribution is not a chisquare, in which case this
    test is not appropriate.

    This function handles masked arrays.  If an element of `f_obs` or `f_exp`
    is masked, then data at that position is ignored, and does not count
    towards the size of the data set.

    .. versionadded:: 0.13.0

    References
    ----------
    .. [1] Lowry, Richard.  "Concepts and Applications of Inferential
           Statistics". Chapter 8.
           https://web.archive.org/web/20171015035606/http://faculty.vassar.edu/lowry/ch8pt1.html
    .. [2] "Chi-squared test", https://en.wikipedia.org/wiki/Chi-squared_test
    .. [3] "G-test", https://en.wikipedia.org/wiki/G-test
    .. [4] Sokal, R. R. and Rohlf, F. J. "Biometry: the principles and
           practice of statistics in biological research", New York: Freeman
           (1981)
    .. [5] Cressie, N. and Read, T. R. C., "Multinomial Goodness-of-Fit
           Tests", J. Royal Stat. Soc. Series B, Vol. 46, No. 3 (1984),
           pp. 440-464.

    Examples
    --------
    (See `chisquare` for more examples.)

    When just `f_obs` is given, it is assumed that the expected frequencies
    are uniform and given by the mean of the observed frequencies.  Here we
    perform a G-test (i.e. use the log-likelihood ratio statistic):

    >>> import mars.tensor as mt
    >>> from mars.tensor.stats import power_divergence
    >>> power_divergence([16, 18, 16, 14, 12, 12], lambda_='log-likelihood').execute()
    (2.006573162632538, 0.84823476779463769)

    The expected frequencies can be given with the `f_exp` argument:

    >>> power_divergence([16, 18, 16, 14, 12, 12],
    ...                  f_exp=[16, 16, 16, 16, 16, 8],
    ...                  lambda_='log-likelihood').execute()
    (3.3281031458963746, 0.6495419288047497)

    When `f_obs` is 2-D, by default the test is applied to each column.

    >>> obs = mt.array([[16, 18, 16, 14, 12, 12], [32, 24, 16, 28, 20, 24]]).T
    >>> obs.shape
    (6, 2)
    >>> power_divergence(obs, lambda_="log-likelihood").execute()
    (array([ 2.00657316,  6.77634498]), array([ 0.84823477,  0.23781225]))

    By setting ``axis=None``, the test is applied to all data in the array,
    which is equivalent to applying the test to the flattened array.

    >>> power_divergence(obs, axis=None).execute()
    (23.31034482758621, 0.015975692534127565)
    >>> power_divergence(obs.ravel()).execute()
    (23.31034482758621, 0.015975692534127565)

    `ddof` is the change to make to the default degrees of freedom.

    >>> power_divergence([16, 18, 16, 14, 12, 12], ddof=1).execute()
    (2.0, 0.73575888234288467)

    `f_obs` and `f_exp` are also broadcast.  In the following, `f_obs` has
    shape (6,) and `f_exp` has shape (2, 6), so the result of broadcasting
    `f_obs` and `f_exp` has shape (2, 6).  To compute the desired chi-squared
    statistics, we must use ``axis=1``:

    >>> power_divergence([16, 18, 16, 14, 12, 12],
    ...                  f_exp=[[16, 16, 16, 16, 16, 8],
    ...                         [8, 20, 20, 16, 12, 12]],
    ...                  axis=1)
    (array([ 3.5 ,  9.25]), array([ 0.62338763,  0.09949846]))
    """
    # Convert the input argument `lambda_` to a numerical value.
    if isinstance(lambda_, str):
        if lambda_ not in _power_div_lambda_names:
            names = repr(list(_power_div_lambda_names.keys()))[1:-1]
            raise ValueError("invalid string for lambda_: {0!r}.  Valid strings "
                             "are {1}".format(lambda_, names))
        lambda_ = _power_div_lambda_names[lambda_]
    elif lambda_ is None:
        lambda_ = 1

    f_obs = asarray(f_obs)

    if f_exp is not None:
        f_exp = asarray(f_exp)
    else:
        f_exp = f_obs.mean(axis=axis, keepdims=True)

    # `terms` is the array of terms that are summed along `axis` to create
    # the test statistic.  We use some specialized code for a few special
    # cases of lambda_.
    if lambda_ == 1:
        # Pearson's chi-squared statistic
        terms = (f_obs.astype(np.float64) - f_exp)**2 / f_exp
    elif lambda_ == 0:
        # Log-likelihood ratio (i.e. G-test)
        terms = 2.0 * special.xlogy(f_obs, f_obs / f_exp)
    elif lambda_ == -1:
        # Modified log-likelihood ratio
        terms = 2.0 * special.xlogy(f_exp, f_exp / f_obs)
    else:
        # General Cressie-Read power divergence.
        terms = f_obs * ((f_obs / f_exp)**lambda_ - 1)
        terms /= 0.5 * lambda_ * (lambda_ + 1)

    stat = terms.sum(axis=axis)

    num_obs = _count(terms, axis=axis)
    # we decide not to support ddof for multiple dimensions
    # ddof = asarray(ddof)
    p = stat.map_chunk(sp_distributions.chi2.sf, (num_obs - 1 - ddof,), elementwise=True)

    return ExecutableTuple(Power_divergenceResult(stat, p))



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...config import options
from ...core import OutputType
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameCount(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = OperandDef.COUNT
    _func_name = 'count'

    @property
    def is_atomic(self):
        return True

    @classmethod
    def get_reduction_callable(cls, op):
        skipna, numeric_only = op.skipna, op.numeric_only

        def count(value):
            if value.ndim == 1:
                return value.count()
            return value.count(skipna=skipna, numeric_only=numeric_only)

        return count


def count_series(series, level=None, combine_size=None, **kw):
    use_inf_as_na = kw.pop('_use_inf_as_na', options.dataframe.mode.use_inf_as_na)
    op = DataFrameCount(level=level, combine_size=combine_size, output_types=[OutputType.scalar],
                        use_inf_as_na=use_inf_as_na)
    return op(series)


def count_dataframe(df, axis=0, level=None, numeric_only=False, combine_size=None, **kw):
    use_inf_as_na = kw.pop('_use_inf_as_na', options.dataframe.mode.use_inf_as_na)
    op = DataFrameCount(axis=axis, level=level, numeric_only=numeric_only, combine_size=combine_size,
                        output_types=[OutputType.series], use_inf_as_na=use_inf_as_na)
    return op(df)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
import os
import time
from collections import OrderedDict

from ... import promise
from ...config import options
from ...utils import parse_readable_size, log_unhandled, readable_size, tokenize
from ...errors import SpillNotConfigured, SpillSizeExceeded, NoDataToSpill, PinDataKeyFailed
from ..utils import WorkerActor
from .core import DataStorageDevice
from .manager import StorageManagerActor

logger = logging.getLogger(__name__)


class ObjectHolderActor(WorkerActor):
    _storage_device = None
    _spill_devices = None

    def __init__(self, size_limit=0):
        super().__init__()
        self._size_limit = size_limit

        self._data_holder = OrderedDict()
        self._data_sizes = dict()

        self._total_hold = 0
        self._pinned_counter = dict()
        self._spill_pending_keys = set()

        self._total_spill = 0
        self._min_spill_size = 0
        self._max_spill_size = 0

        self._dispatch_ref = None
        self._status_ref = None
        self._storage_handler = None

    def post_create(self):
        from ..dispatcher import DispatchActor
        from ..status import StatusActor

        super().post_create()
        self.register_actors_down_handler()
        self._dispatch_ref = self.promise_ref(DispatchActor.default_uid())

        parse_num, is_percent = parse_readable_size(options.worker.min_spill_size)
        self._min_spill_size = int(self._size_limit * parse_num if is_percent else parse_num)
        parse_num, is_percent = parse_readable_size(options.worker.max_spill_size)
        self._max_spill_size = int(self._size_limit * parse_num if is_percent else parse_num)

        status_ref = self.ctx.actor_ref(StatusActor.default_uid())
        self._status_ref = status_ref if self.ctx.has_actor(status_ref) else None

        self._storage_handler = self.storage_client.get_storage_handler(
            self._storage_device.build_location(self.proc_id))

        self.ref().update_cache_status(_tell=True)

    def pre_destroy(self):
        for k in self._data_holder:
            self._data_holder[k] = None

    def update_cache_status(self):
        raise NotImplementedError

    def post_delete(self, session_id, data_keys):
        raise NotImplementedError

    def get_size_limit(self):
        return self._size_limit

    @promise.reject_on_exception
    @log_unhandled
    def spill_size(self, size, multiplier=1, callback=None):
        if not self._spill_devices:  # pragma: no cover
            raise SpillNotConfigured

        request_size = int(size * multiplier)
        request_size = max(request_size, self._min_spill_size)
        if request_size > self._size_limit:
            raise SpillSizeExceeded
        request_size = min(request_size, self._max_spill_size)

        spill_ref_key = tokenize((time.time(), size, multiplier))

        logger.debug('Start spilling %d(x%d) bytes in %s. ref_key==%s.',
                     request_size, multiplier, self.uid, spill_ref_key)

        if request_size + self._total_hold > self._size_limit:
            acc_free = 0
            free_keys = []
            for k in self._data_holder.keys():
                if k in self._pinned_counter or k in self._spill_pending_keys:
                    continue
                acc_free += self._data_sizes[k]
                free_keys.append(k)
                self._spill_pending_keys.add(k)
                if request_size + self._total_hold - acc_free <= self._size_limit:
                    break

            if not free_keys:
                logger.warning('Cannot spill further in %s. Rejected. request=%d ref_key=%s',
                               self.uid, request_size, spill_ref_key)
                raise NoDataToSpill

            logger.debug('Decide to spill %d data keys %r in %s. request=%d ref_key=%s',
                         len(free_keys), free_keys, self.uid, request_size, spill_ref_key)

            @log_unhandled
            def _release_spill_allocations(key):
                logger.debug('Removing reference of data %s from %s when spilling. ref_key=%s',
                             key, self.uid, spill_ref_key)
                self.delete_objects(key[0], [key[1]])

            @log_unhandled
            def _handle_spill_reject(*exc, **kwargs):
                key = kwargs['session_data_key']
                self._remove_spill_pending(*key)
                raise exc[1].with_traceback(exc[2])

            @log_unhandled
            def _spill_key(key):
                if key in self._pinned_counter or key not in self._data_holder:
                    self._remove_spill_pending(*key)
                    return
                logger.debug('Spilling key %s in %s. ref_key=%s', key, self.uid, spill_ref_key)
                return self.storage_client.copy_to(key[0], [key[1]], self._spill_devices) \
                    .then(lambda *_: _release_spill_allocations(key),
                          functools.partial(_handle_spill_reject, session_data_key=key))

            @log_unhandled
            def _finalize_spill(*_):
                logger.debug('Finish spilling %d data keys in %s. ref_key=%s',
                             len(free_keys), self.uid, spill_ref_key)
                self._plasma_client.evict(request_size)
                if callback:
                    self.tell_promise(callback)
                self.update_cache_status()

            promise.all_(_spill_key(k) for k in free_keys).then(_finalize_spill) \
                .catch(lambda *exc: self.tell_promise(callback, *exc, _accept=False))
        else:
            logger.debug('No need to spill in %s. request=%d ref_key=%s',
                         self.uid, request_size, spill_ref_key)

            self._plasma_client.evict(request_size)
            if callback:
                self.tell_promise(callback)

    @log_unhandled
    def _internal_put_object(self, session_id, data_key, obj, size):
        try:
            session_data_key = (session_id, data_key)
            if session_data_key in self._data_holder:
                self._total_hold -= self._data_sizes[session_data_key]
                del self._data_holder[session_data_key]

            self._data_holder[session_data_key] = obj
            self._data_sizes[session_data_key] = size
            self._total_hold += size
        finally:
            del obj

    def _finish_put_objects(self, _session_id, data_keys):
        if logger.getEffectiveLevel() <= logging.DEBUG:  # pragma: no cover
            simplified_keys = sorted(set(k[0] if isinstance(k, tuple) else k for k in data_keys))
            logger.debug('Data %r registered in %s. total_hold=%d', simplified_keys,
                         self.uid, self._total_hold)
        self.update_cache_status()

    def _remove_spill_pending(self, session_id, data_key):
        try:
            self._spill_pending_keys.remove((session_id, data_key))
            logger.debug('Spill-pending key (%s, %s) removed in %s', session_id, data_key, self.uid)
        except KeyError:
            pass

    @log_unhandled
    def delete_objects(self, session_id, data_keys):
        actual_removed = []
        for data_key in data_keys:
            session_data_key = (session_id, data_key)

            self._remove_spill_pending(session_id, data_key)

            try:
                del self._pinned_counter[session_data_key]
            except KeyError:
                pass

            if session_data_key in self._data_holder:
                actual_removed.append(data_key)

                data_size = self._data_sizes[session_data_key]
                self._total_hold -= data_size
                del self._data_holder[session_data_key]
                del self._data_sizes[session_data_key]

        self.post_delete(session_id, actual_removed)
        if actual_removed:
            logger.debug('Data %s unregistered in %s. total_hold=%d', actual_removed, self.uid, self._total_hold)
            self.update_cache_status()

    def lift_data_keys(self, session_id, data_keys, last=True):
        for k in data_keys:
            self._data_holder.move_to_end((session_id, k), last)

    @log_unhandled
    def pin_data_keys(self, session_id, data_keys, token):
        spilling_keys = list(k for k in data_keys if (session_id, k) in self._spill_pending_keys)
        if spilling_keys:
            logger.warning('Cannot pin data key %r: under spilling', spilling_keys)
            raise PinDataKeyFailed
        pinned = []
        for k in data_keys:
            session_k = (session_id, k)
            if session_k not in self._data_holder:
                continue
            if session_k not in self._pinned_counter:
                self._pinned_counter[session_k] = set()
            self._pinned_counter[session_k].add(token)
            pinned.append(k)
        logger.debug('Data keys %r pinned in %s', pinned, self.uid)
        return pinned

    @log_unhandled
    def unpin_data_keys(self, session_id, data_keys, token):
        unpinned = []
        for k in data_keys:
            session_k = (session_id, k)
            try:
                self._pinned_counter[session_k].remove(token)
                if not self._pinned_counter[session_k]:
                    del self._pinned_counter[session_k]
                unpinned.append(k)
            except KeyError:
                continue
        if unpinned:
            logger.debug('Data keys %r unpinned in %s', unpinned, self.uid)
        return unpinned

    def dump_keys(self):  # pragma: no cover
        return list(self._data_holder.keys())


class SimpleObjectHolderActor(ObjectHolderActor):
    def post_create(self):
        super().post_create()
        manager_ref = self.ctx.actor_ref(StorageManagerActor.default_uid())
        manager_ref.register_process_holder(
            self.proc_id, self._storage_device, self.ref())

    def put_objects(self, session_id, data_keys, data_objs, data_sizes, pin_token=None):
        try:
            for data_key, obj, size in zip(data_keys, data_objs, data_sizes):
                self._internal_put_object(session_id, data_key, obj, size)
            if pin_token:
                self.pin_data_keys(session_id, data_keys, pin_token)
            self._finish_put_objects(session_id, data_keys)
        finally:
            del data_objs

    def get_object(self, session_id, data_key):
        return self._data_holder[(session_id, data_key)]

    def get_objects(self, session_id, data_keys):
        return [self._data_holder[(session_id, key)] for key in data_keys]

    def update_cache_status(self):
        pass

    def post_delete(self, session_id, data_keys):
        pass


class SharedHolderActor(ObjectHolderActor):
    if options.vineyard.enabled:
        _storage_device = DataStorageDevice.VINEYARD  # pragma: no cover
    else:
        _storage_device = DataStorageDevice.SHARED_MEMORY
    _spill_devices = (DataStorageDevice.DISK,)

    def post_create(self):
        super().post_create()
        if not options.vineyard.enabled:
            self._size_limit = self._shared_store.get_actual_capacity(self._size_limit)
        logger.info('Detected actual plasma store size: %s', readable_size(self._size_limit))

    def update_cache_status(self):
        if self._status_ref:
            self._status_ref.set_cache_allocations(
                dict(hold=self._total_hold, total=self._size_limit), _tell=True, _wait=False)

    def post_delete(self, session_id, data_keys):
        if self._shared_store:
            self._shared_store.batch_delete(session_id, data_keys)
        self._storage_handler.unregister_data(session_id, data_keys)

    def put_objects_by_keys(self, session_id, data_keys, shapes=None, pin_token=None):
        sizes = []
        for data_key in data_keys:
            buf = None
            try:
                buf = self._shared_store.get_buffer(session_id, data_key)
                size = len(buf)
                self._internal_put_object(session_id, data_key, buf, size)
            finally:
                del buf
            sizes.append(size)
        if pin_token:
            self.pin_data_keys(session_id, data_keys, pin_token)

        self._finish_put_objects(session_id, data_keys)
        self.storage_client.register_data(
            session_id, data_keys, (0, self._storage_device), sizes, shapes=shapes)


class InProcHolderActor(SimpleObjectHolderActor):
    _storage_device = DataStorageDevice.PROC_MEMORY


class CudaHolderActor(SimpleObjectHolderActor):
    _storage_device = DataStorageDevice.CUDA
    if options.vineyard.enabled:
        shared_memory_device = DataStorageDevice.VINEYARD  # pragma: no cover
    else:
        shared_memory_device = DataStorageDevice.SHARED_MEMORY

    _spill_devices = [shared_memory_device, DataStorageDevice.DISK]

    def __init__(self, size_limit=0, device_id=None):
        super().__init__(size_limit=size_limit)
        if device_id is not None:
            os.environ['CUDA_VISIBLE_DEVICES'] = str(device_id)

        # warm up cupy
        try:
            import cupy
            cupy.zeros((10, 10)).sum()
        except ImportError:
            pass
        # warm up cudf
        try:
            import cudf
            import numpy as np
            import pandas as pd
            cudf.from_pandas(pd.DataFrame(np.zeros((10, 10))))
        except ImportError:
            pass



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from .. import resource
from ..config import options
from ..base_app import BaseApplication
from ..distributor import MarsDistributor
from .service import SchedulerService

logger = logging.getLogger(__name__)


class SchedulerApplication(BaseApplication):
    service_description = 'Mars Scheduler'
    service_logger = logger

    def __init__(self):
        super().__init__()
        self._service = None

    def config_args(self, parser):
        super().config_args(parser)
        parser.add_argument('--nproc', help='number of processes')
        parser.add_argument('--disable-failover', action='store_true',
                            help='disable fail-over')

    def parse_args(self, parser, argv, environ=None):
        args = super().parse_args(parser, argv)
        environ = environ or os.environ
        args.disable_failover = args.disable_failover \
            or bool(int(environ.get('MARS_DISABLE_FAILOVER', '0')))
        options.scheduler.dump_graph_data = bool(int(environ.get('MARS_DUMP_GRAPH_DATA', '0')))
        return args

    def create_pool(self, *args, **kwargs):
        self._service = SchedulerService(disable_failover=self.args.disable_failover)
        self.n_process = int(self.args.nproc or resource.cpu_count())
        kwargs['distributor'] = MarsDistributor(self.n_process, 's:h1:')
        return super().create_pool(*args, **kwargs)

    def create_scheduler_discoverer(self):
        advertise_endpoint = self.args.advertise or self.endpoint
        if ':' not in advertise_endpoint:
            advertise_endpoint += ':' + self.endpoint.rsplit(':', 1)[-1]
        all_schedulers = {advertise_endpoint}

        if self.args.schedulers:
            all_schedulers.update(self.args.schedulers.split(','))
        self.args.schedulers = ','.join(all_schedulers)

        super().create_scheduler_discoverer()

    def start(self):
        self._service.start(self.endpoint, self.scheduler_discoverer, self.pool)

    def stop(self):
        self._service.stop(self.pool)


main = SchedulerApplication()

if __name__ == '__main__':
    main()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from collections.abc import Iterable
from distutils.version import LooseVersion

import cloudpickle
import numpy as np
import pandas as pd
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy

_HAS_SQUEEZE = LooseVersion(pd.__version__) < '1.1.0'


class GroupByWrapper:
    def __init__(self, obj, groupby_obj=None, keys=None, axis=0, level=None, grouper=None,
                 exclusions=None, selection=None, as_index=True, sort=True,
                 group_keys=True, squeeze=False, observed=False, mutated=False,
                 grouper_cache=None):

        def fill_value(v, key):
            return v if v is not None or groupby_obj is None else getattr(groupby_obj, key)

        self.obj = obj
        self.keys = fill_value(keys, 'keys')
        self.axis = fill_value(axis, 'axis')
        self.level = fill_value(level, 'level')
        self.exclusions = fill_value(exclusions, 'exclusions')
        self.selection = selection
        self.as_index = fill_value(as_index, 'as_index')
        self.sort = fill_value(sort, 'sort')
        self.group_keys = fill_value(group_keys, 'group_keys')
        self.squeeze = fill_value(squeeze, 'squeeze')
        self.observed = fill_value(observed, 'observed')
        self.mutated = fill_value(mutated, 'mutated')

        if groupby_obj is None:
            groupby_kw = dict(keys=keys, axis=axis, level=level, grouper=grouper,
                              exclusions=exclusions, as_index=as_index, group_keys=group_keys,
                              squeeze=squeeze, observed=observed, mutated=mutated)
            if not _HAS_SQUEEZE:  # pragma: no branch
                groupby_kw.pop('squeeze')

            if obj.ndim == 2:
                self.groupby_obj = DataFrameGroupBy(obj, **groupby_kw)
            else:
                self.groupby_obj = SeriesGroupBy(obj, **groupby_kw)
        else:
            self.groupby_obj = groupby_obj

        if grouper_cache:
            self.groupby_obj.grouper._cache = grouper_cache
        if selection:
            self.groupby_obj = self.groupby_obj[selection]

        self.is_frame = isinstance(self.groupby_obj, DataFrameGroupBy)

    def __getitem__(self, item):
        return GroupByWrapper(
            self.obj, keys=self.keys, axis=self.axis, level=self.level,
            grouper=self.groupby_obj.grouper, exclusions=self.exclusions, selection=item,
            as_index=self.as_index, sort=self.sort, group_keys=self.group_keys,
            squeeze=self.squeeze, observed=self.observed, mutated=self.mutated)

    def __getattr__(self, item):
        if item.startswith('_'):  # pragma: no cover
            return object.__getattribute__(self, item)
        if item in getattr(self.obj, 'columns', ()):
            return self.__getitem__(item)
        return getattr(self.groupby_obj, item)

    def __iter__(self):
        return self.groupby_obj.__iter__()

    def __sizeof__(self):
        return sys.getsizeof(self.obj) \
            + sys.getsizeof(getattr(self.groupby_obj.grouper, '_cache', None))

    def __reduce__(self):
        return type(self).from_tuple, (self.to_tuple(pickle_function=True, truncate=True),)

    def __bool__(self):
        return bool(np.prod(self.shape))

    @property
    def empty(self):
        return self.obj.empty

    @property
    def shape(self):
        shape = list(self.groupby_obj.obj.shape)
        if self.is_frame and self.selection:
            shape[1] = len(self.selection)
        return tuple(shape)

    @property
    def _selected_obj(self):
        return getattr(self.groupby_obj, '_selected_obj')

    def to_tuple(self, truncate=False, pickle_function=False):
        if self.selection and truncate:
            if isinstance(self.selection, Iterable) and not isinstance(self.selection, str):
                item_list = list(self.selection)
            else:
                item_list = [self.selection]
            item_set = set(item_list)

            if isinstance(self.keys, list):
                sel_keys = self.keys
            elif self.keys in self.obj.columns:
                sel_keys = [self.keys]
            else:
                sel_keys = []

            all_items = item_list + [k for k in sel_keys or () if k not in item_set]
            if set(all_items) == set(self.obj.columns):
                obj = self.obj
            else:
                obj = self.obj[all_items]
        else:
            obj = self.obj

        if pickle_function and callable(self.keys):
            keys = cloudpickle.dumps(self.keys)
        else:
            keys = self.keys

        return obj, keys, self.axis, self.level, self.exclusions, self.selection, \
            self.as_index, self.sort, self.group_keys, self.squeeze, self.observed, \
            self.mutated, getattr(self.groupby_obj.grouper, '_cache', dict())

    @classmethod
    def from_tuple(cls, tp):
        obj, keys, axis, level, exclusions, selection, as_index, sort, group_keys, squeeze, \
            observed, mutated, grouper_cache = tp

        if isinstance(keys, (bytes, bytearray)):
            keys = cloudpickle.loads(keys)

        return cls(obj, keys=keys, axis=axis, level=level, exclusions=exclusions, selection=selection,
                   as_index=as_index, sort=sort, group_keys=group_keys, squeeze=squeeze, observed=observed,
                   mutated=mutated, grouper_cache=grouper_cache)


def wrapped_groupby(obj, by=None, axis=0, level=None, as_index=True, sort=True, group_keys=True,
                    squeeze=False, observed=False):
    groupby_kw = dict(by=by, axis=axis, level=level, as_index=as_index, sort=sort,
                      group_keys=group_keys, squeeze=squeeze, observed=observed)
    if not _HAS_SQUEEZE:  # pragma: no branch
        groupby_kw.pop('squeeze')

    groupby_obj = obj.groupby(**groupby_kw)
    return GroupByWrapper(obj, groupby_obj=groupby_obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..datasource import tensor as astensor
from .split import split


def dsplit(a, indices_or_sections):
    """
    Split tensor into multiple sub-tensors along the 3rd axis (depth).

    Please refer to the `split` documentation.  `dsplit` is equivalent
    to `split` with ``axis=2``, the array is always split along the third
    axis provided the tensor dimension is greater than or equal to 3.

    See Also
    --------
    split : Split a tensor into multiple sub-arrays of equal size.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(16.0).reshape(2, 2, 4)
    >>> x.execute()
    array([[[  0.,   1.,   2.,   3.],
            [  4.,   5.,   6.,   7.]],
           [[  8.,   9.,  10.,  11.],
            [ 12.,  13.,  14.,  15.]]])
    >>> mt.dsplit(x, 2).execute()
    [array([[[  0.,   1.],
            [  4.,   5.]],
           [[  8.,   9.],
            [ 12.,  13.]]]),
     array([[[  2.,   3.],
            [  6.,   7.]],
           [[ 10.,  11.],
            [ 14.,  15.]]])]
    >>> mt.dsplit(x, mt.array([3, 6])).execute()
    [array([[[  0.,   1.,   2.],
            [  4.,   5.,   6.]],
           [[  8.,   9.,  10.],
            [ 12.,  13.,  14.]]]),
     array([[[  3.],
            [  7.]],
           [[ 11.],
            [ 15.]]]),
     array([], dtype=float64)]

    """
    ary = a
    a = astensor(a)

    if a.ndim < 3:
        raise ValueError('dsplit only works on tensors of 3 or more dimensions')
    return split(ary, indices_or_sections, 2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode='binary_or')
class TensorMod(TensorBinOp):
    _op_type_ = OperandDef.MOD
    _func_name = 'mod'


@infer_dtype(np.mod)
def mod(x1, x2, out=None, where=None, **kwargs):
    """
    Return element-wise remainder of division.

    Computes the remainder complementary to the `floor_divide` function.  It is
    equivalent to the Python modulus operator``x1 % x2`` and has the same sign
    as the divisor `x2`. The MATLAB function equivalent to ``np.remainder``
    is ``mod``.

    .. warning::

        This should not be confused with:

        * Python 3.7's `math.remainder` and C's ``remainder``, which
          computes the IEEE remainder, which are the complement to
          ``round(x1 / x2)``.
        * The MATLAB ``rem`` function and or the C ``%`` operator which is the
          complement to ``int(x1 / x2)``.

    Parameters
    ----------
    x1 : array_like
        Dividend array.
    x2 : array_like
        Divisor array.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    y : Tensor
        The element-wise remainder of the quotient ``floor_divide(x1, x2)``.
        Returns a scalar if both  `x1` and `x2` are scalars.

    See Also
    --------
    floor_divide : Equivalent of Python ``//`` operator.
    divmod : Simultaneous floor division and remainder.
    fmod : Equivalent of the MATLAB ``rem`` function.
    divide, floor

    Notes
    -----
    Returns 0 when `x2` is 0 and both `x1` and `x2` are (tensors of)
    integers.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.remainder([4, 7], [2, 3]).execute()
    array([0, 1])
    >>> mt.remainder(mt.arange(7), 5).execute()
    array([0, 1, 2, 3, 4, 0, 1])
    """
    op = TensorMod(**kwargs)
    return op(x1, x2, out=out, where=where)


remainder = mod


@infer_dtype(np.mod, reverse=True)
def rmod(x1, x2, **kwargs):
    op = TensorMod(**kwargs)
    return op.rcall(x1, x2)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import defaultdict

import psutil

from .utils import WorkerActor
from ..utils import kill_process_tree

logger = logging.getLogger(__name__)


class WorkerDaemonActor(WorkerActor):
    """
    Actor handling termination of worker processes. If a process terminates
    unprecedentedly, this actor restarts actors previously started
    """
    def __init__(self):
        super().__init__()
        self._proc_actors = defaultdict(dict)
        self._proc_pids = dict()
        self._killed_pids = set()

        self._actor_callbacks = []
        self._proc_callbacks = []

    def create_actor(self, *args, **kwargs):
        """
        Create an actor and record its creation args for recovery
        """
        ref = self.ctx.create_actor(*args, **kwargs)
        proc_idx = self.ctx.distributor.distribute(ref.uid)
        ref_key = (ref.uid, ref.address)
        self._proc_actors[proc_idx][ref_key] = (ref_key, args, kwargs, False)
        return ref

    def destroy_actor(self, ref):
        proc_idx = self.ctx.distributor.distribute(ref.uid)
        ref_key = (ref.uid, ref.address)
        self._proc_actors[proc_idx].pop(ref_key, None)
        self.ctx.destroy_actor(ref)

    def register_child_actor(self, actor_ref):
        proc_idx = self.ctx.distributor.distribute(actor_ref.uid)
        ref_key = (actor_ref.uid, actor_ref.address)
        self._proc_actors[proc_idx][ref_key] = (ref_key, None, None, True)

    def register_process(self, actor_ref, pid):
        """
        Register a process id
        """
        proc_idx = self.ctx.distributor.distribute(actor_ref.uid)
        self._proc_pids[proc_idx] = pid

    def register_actor_callback(self, actor_ref, func):
        """
        Register a callback on an actor for handling process down,
        dead actors will be passed

        :param actor_ref: ActorRef to handle the callback
        :param func: function name of the callback
        """
        self._actor_callbacks.append((actor_ref.uid, actor_ref.address, func))

    def register_process_callback(self, actor_ref, func):
        """
        Register a callback on an actor for handling process down,
        indices of dead processes will be passed

        :param actor_ref: ActorRef to handle the callback
        :param func: function name of the callback
        """
        self._proc_callbacks.append((actor_ref.uid, actor_ref.address, func))

    def kill_actor_process(self, actor_ref):
        """
        Kill a process given the ref of an actor on it
        """
        proc_idx = self.ctx.distributor.distribute(actor_ref.uid)
        try:
            pid = self._proc_pids[proc_idx]
            self._killed_pids.add(pid)
            kill_process_tree(pid)
        except (KeyError, OSError):
            pass

    def is_actor_process_alive(self, actor_ref):
        """
        Check if the process holding the actor is still alive
        """
        proc_idx = self.ctx.distributor.distribute(actor_ref.uid)
        pid = self._proc_pids[proc_idx]
        if pid in self._killed_pids:
            return False
        return psutil.pid_exists(pid)

    def handle_process_down(self, proc_indices):
        """
        When process down is detected,
        :param proc_indices: indices of processes in Mars Worker
        """
        # invoke registered callbacks for processes
        for cb in self._proc_callbacks:
            uid, addr, func = cb
            ref = self.ctx.actor_ref(uid, address=addr)
            getattr(ref, func)(proc_indices, _tell=True)

        # recreate actors given previous records
        refs = set()
        for proc_idx in proc_indices:
            for actor_args in self._proc_actors[proc_idx].values():
                ref_key, args, kw, is_child = actor_args
                refs.add(ref_key)
                if not is_child:
                    self.ctx.create_actor(*args, **kw)

        # invoke registered callbacks for actors
        for cb in self._actor_callbacks:
            uid, addr, func = cb
            ref = self.ctx.actor_ref(uid, address=addr)
            clean_refs = [self.ctx.actor_ref(u, address=None) for u, _ in refs] \
                + [self.ctx.actor_ref(u, address=a) for u, a in refs]
            getattr(ref, func)(clean_refs, _tell=True)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any

import numpy as np
import pandas as pd
from pandas.api.types import is_scalar, is_dict_like

from ... import opcodes
from ...serialize import KeyField, StringField, BoolField, AnyField
from ...tensor import tensor as astensor
from ...tensor.core import TENSOR_CHUNK_TYPE
from ...tiles import TilesError
from ..core import DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE, INDEX_CHUNK_TYPE
from ..initializer import DataFrame as asdataframe, Series as asseries, Index as asindex
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index


class DataFrameToDatetime(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.TO_DATETIME

    _arg = KeyField('arg')
    _errors = StringField('errors')
    _dayfirst = BoolField('dayfirst')
    _yearfirst = BoolField('yearfirst')
    _utc = BoolField('utc')
    _format = StringField('format')
    _exact = BoolField('exact')
    _unit = StringField('unit')
    _infer_datetime_format = BoolField('infer_datetime_format')
    _origin = AnyField('origin')
    _cache = BoolField('cache')

    def __init__(self, errors=None, dayfirst=None, yearfirst=None, utc=None,
                 format=None, exact=None, unit=None, infer_datetime_format=None,
                 origin=None, cache=None, **kw):
        super().__init__(_errors=errors, _dayfirst=dayfirst, _yearfirst=yearfirst,
                         _utc=utc, _format=format, _exact=exact, _unit=unit,
                         _infer_datetime_format=infer_datetime_format, _origin=origin,
                         _cache=cache, **kw)

    @property
    def arg(self):
        return self._arg

    @property
    def errors(self):
        return self._errors

    @property
    def dayfirst(self):
        return self._dayfirst

    @property
    def yearfirst(self):
        return self._yearfirst

    @property
    def utc(self):
        return self._utc

    @property
    def format(self):
        return self._format

    @property
    def exact(self):
        return self._exact

    @property
    def unit(self):
        return self._unit

    @property
    def infer_datetime_format(self):
        return self._infer_datetime_format

    @property
    def origin(self):
        return self._origin

    @property
    def cache(self):
        return self._cache

    @property
    def _params(self):
        return tuple(getattr(self, k) for k in self._keys_
                     if k not in self._no_copy_attrs_ and
                     k != '_arg' and hasattr(self, k))

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._arg = self._inputs[0]

    def __call__(self, arg):
        if is_scalar(arg):
            ret = pd.to_datetime(arg, errors=self._errors, dayfirst=self._dayfirst,
                                 yearfirst=self._yearfirst, utc=self._utc,
                                 format=self._format, exact=self._exact,
                                 unit=self._unit, infer_datetime_format=self._infer_datetime_format,
                                 origin=self._origin, cache=self._cache)
            return astensor(ret)

        dtype = np.datetime64(1, 'ns').dtype
        if isinstance(arg, (pd.Series, SERIES_TYPE)):
            arg = asseries(arg)
            return self.new_series([arg], shape=arg.shape,
                                   dtype=dtype, index_value=arg.index_value,
                                   name=arg.name)
        if is_dict_like(arg) or isinstance(arg, DATAFRAME_TYPE):
            arg = asdataframe(arg)
            columns = arg.columns_value.to_pandas().tolist()
            if sorted(columns) != sorted(['year', 'month', 'day']):
                missing = ','.join(c for c in ['day', 'month', 'year'] if c not in columns)
                raise ValueError('to assemble mappings requires at least '
                                 f'that [year, month, day] be specified: [{missing}] is missing')
            return self.new_series([arg], shape=(arg.shape[0],),
                                   dtype=dtype, index_value=arg.index_value)
        elif isinstance(arg, (pd.Index, INDEX_TYPE)):
            arg = asindex(arg)
            return self.new_index([arg], shape=arg.shape,
                                  dtype=dtype,
                                  index_value=parse_index(pd.Index([], dtype=dtype),
                                                          self._params, arg),
                                  name=arg.name)
        else:
            arg = astensor(arg)
            if arg.ndim != 1:
                raise TypeError('arg must be a string, datetime, '
                                'list, tuple, 1-d tensor, or Series')
            return self.new_index([arg], shape=arg.shape,
                                  dtype=dtype,
                                  index_value=parse_index(pd.Index([], dtype=dtype),
                                                          self._params, arg))

    @classmethod
    def tile(cls, op: "DataFrameToDatetime"):
        out = op.outputs[0]
        arg = op.arg

        if isinstance(arg, DATAFRAME_TYPE):
            if np.isnan(arg.shape[0]) or \
                    any(np.isnan(s) for s in arg.nsplits[1]):  # pragma: no cover
                raise TilesError('unknown chunk shape on columns axis')

            arg = arg.rechunk({1: arg.shape[1]})._inplace_tile()

        out_chunks = []
        for chunk in arg.chunks:
            chunk_op = op.copy().reset_key()
            if isinstance(chunk, (TENSOR_CHUNK_TYPE, INDEX_CHUNK_TYPE)):
                chunk_index_value = parse_index(
                    pd.Index([], dtype=out.dtype), op._params, chunk)
            else:
                chunk_index_value = chunk.index_value

            out_chunk = chunk_op.new_chunk([chunk], shape=(chunk.shape[0],),
                                           dtype=out.dtype,
                                           index_value=chunk_index_value,
                                           name=out.name,
                                           index=(chunk.index[0],))
            out_chunks.append(out_chunk)

        params = out.params
        params['nsplits'] = (arg.nsplits[0],)
        params['chunks'] = out_chunks
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op: "DataFrameToDatetime"):
        arg = ctx[op.arg.key]

        call = partial(pd.to_datetime,
                       errors=op.errors, dayfirst=op.dayfirst,
                       yearfirst=op.yearfirst, utc=op.utc,
                       format=op.format, exact=op.exact, unit=op.unit,
                       infer_datetime_format=op.infer_datetime_format,
                       origin=op.origin, cache=op.cache)

        try:
            ctx[op.outputs[0].key] = call(arg)
        except ValueError:  # pragma: no cover
            ctx[op.outputs[0].key] = call(arg.copy())


def to_datetime(arg,
                errors: str = 'raise',
                dayfirst: bool = False,
                yearfirst: bool = False,
                utc: bool = None,
                format: str = None,
                exact: bool = True,
                unit: str = None,
                infer_datetime_format: bool = False,
                origin: Any = 'unix',
                cache: bool = True):
    """
    Convert argument to datetime.

    Parameters
    ----------
    arg : int, float, str, datetime, list, tuple, 1-d array, Series DataFrame/dict-like
        The object to convert to a datetime.
    errors : {'ignore', 'raise', 'coerce'}, default 'raise'
        - If 'raise', then invalid parsing will raise an exception.
        - If 'coerce', then invalid parsing will be set as NaT.
        - If 'ignore', then invalid parsing will return the input.
    dayfirst : bool, default False
        Specify a date parse order if `arg` is str or its list-likes.
        If True, parses dates with the day first, eg 10/11/12 is parsed as
        2012-11-10.
        Warning: dayfirst=True is not strict, but will prefer to parse
        with day first (this is a known bug, based on dateutil behavior).
    yearfirst : bool, default False
        Specify a date parse order if `arg` is str or its list-likes.

        - If True parses dates with the year first, eg 10/11/12 is parsed as
          2010-11-12.
        - If both dayfirst and yearfirst are True, yearfirst is preceded (same
          as dateutil).

        Warning: yearfirst=True is not strict, but will prefer to parse
        with year first (this is a known bug, based on dateutil behavior).
    utc : bool, default None
        Return UTC DatetimeIndex if True (converting any tz-aware
        datetime.datetime objects as well).
    format : str, default None
        The strftime to parse time, eg "%d/%m/%Y", note that "%f" will parse
        all the way up to nanoseconds.
        See strftime documentation for more information on choices:
        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior.
    exact : bool, True by default
        Behaves as:
        - If True, require an exact format match.
        - If False, allow the format to match anywhere in the target string.

    unit : str, default 'ns'
        The unit of the arg (D,s,ms,us,ns) denote the unit, which is an
        integer or float number. This will be based off the origin.
        Example, with unit='ms' and origin='unix' (the default), this
        would calculate the number of milliseconds to the unix epoch start.
    infer_datetime_format : bool, default False
        If True and no `format` is given, attempt to infer the format of the
        datetime strings, and if it can be inferred, switch to a faster
        method of parsing them. In some cases this can increase the parsing
        speed by ~5-10x.
    origin : scalar, default 'unix'
        Define the reference date. The numeric values would be parsed as number
        of units (defined by `unit`) since this reference date.

        - If 'unix' (or POSIX) time; origin is set to 1970-01-01.
        - If 'julian', unit must be 'D', and origin is set to beginning of
          Julian Calendar. Julian day number 0 is assigned to the day starting
          at noon on January 1, 4713 BC.
        - If Timestamp convertible, origin is set to Timestamp identified by
          origin.
    cache : bool, default True
        If True, use a cache of unique, converted dates to apply the datetime
        conversion. May produce significant speed-up when parsing duplicate
        date strings, especially ones with timezone offsets. The cache is only
        used when there are at least 50 values. The presence of out-of-bounds
        values will render the cache unusable and may slow down parsing.

    Returns
    -------
    datetime
        If parsing succeeded.
        Return type depends on input:

        - list-like: DatetimeIndex
        - Series: Series of datetime64 dtype
        - scalar: Timestamp

        In case when it is not possible to return designated types (e.g. when
        any element of input is before Timestamp.min or after Timestamp.max)
        return will have datetime.datetime type (or corresponding
        array/Series).

    See Also
    --------
    DataFrame.astype : Cast argument to a specified dtype.
    to_timedelta : Convert argument to timedelta.
    convert_dtypes : Convert dtypes.

    Examples
    --------
    Assembling a datetime from multiple columns of a DataFrame. The keys can be
    common abbreviations like ['year', 'month', 'day', 'minute', 'second',
    'ms', 'us', 'ns']) or plurals of the same

    >>> import mars.dataframe as md

    >>> df = md.DataFrame({'year': [2015, 2016],
    ...                    'month': [2, 3],
    ...                    'day': [4, 5]})
    >>> md.to_datetime(df).execute()
    0   2015-02-04
    1   2016-03-05
    dtype: datetime64[ns]

    If a date does not meet the `timestamp limitations
    <https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html
    #timeseries-timestamp-limits>`_, passing errors='ignore'
    will return the original input instead of raising any exception.

    Passing errors='coerce' will force an out-of-bounds date to NaT,
    in addition to forcing non-dates (or non-parseable dates) to NaT.

    >>> md.to_datetime('13000101', format='%Y%m%d', errors='ignore').execute()
    datetime.datetime(1300, 1, 1, 0, 0)
    >>> md.to_datetime('13000101', format='%Y%m%d', errors='coerce').execute()
    NaT

    Passing infer_datetime_format=True can often-times speedup a parsing
    if its not an ISO8601 format exactly, but in a regular format.

    >>> s = md.Series(['3/11/2000', '3/12/2000', '3/13/2000'] * 1000)
    >>> s.head().execute()
    0    3/11/2000
    1    3/12/2000
    2    3/13/2000
    3    3/11/2000
    4    3/12/2000
    dtype: object

    Using a unix epoch time

    >>> md.to_datetime(1490195805, unit='s').execute()
    Timestamp('2017-03-22 15:16:45')
    >>> md.to_datetime(1490195805433502912, unit='ns').execute()
    Timestamp('2017-03-22 15:16:45.433502912')

    .. warning:: For float arg, precision rounding might happen. To prevent
        unexpected behavior use a fixed-width exact type.

    Using a non-unix epoch origin

    >>> md.to_datetime([1, 2, 3], unit='D',
    ...                origin=md.Timestamp('1960-01-01')).execute()
    DatetimeIndex(['1960-01-02', '1960-01-03', '1960-01-04'], \
dtype='datetime64[ns]', freq=None)
    """
    op = DataFrameToDatetime(errors=errors, dayfirst=dayfirst,
                             yearfirst=yearfirst, utc=utc, format=format,
                             exact=exact, unit=unit,
                             infer_datetime_format=infer_datetime_format,
                             origin=origin, cache=cache)
    return op(arg)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from bokeh.models import ColumnDataSource, Legend
from bokeh.plotting import figure

from .server import MarsWebAPI, MarsRequestHandler, get_jinja_env, \
    register_bokeh_app, register_web_handler
from ..actors import new_client, ActorNotExist
from ..scheduler import OperandState
from ..utils import to_str


_actor_client = new_client()
_jinja_env = get_jinja_env()

PROGRESS_APP_NAME = 'bk_task_progress'
_base_palette = ['#e0f9ff', '#81d4dd', '#ff9e3b', '#399e34', '#2679b2', '#b3de8e',
                 '#e01f27', '#b3b3cc', '#f0f0f5']


class TaskListHandler(MarsRequestHandler):
    def get(self):
        sessions = self.web_api.get_tasks_info()

        template = _jinja_env.get_template('task_pages/list.html')
        self.write_rendered(template, sessions=sessions)


class TaskHandler(MarsRequestHandler):
    def get(self, session_id, graph_key):
        session_name = session_id

        template = _jinja_env.get_template('task_pages/progress.html')
        task_progress_script = self.bokeh_server_document(
            PROGRESS_APP_NAME, arguments=dict(session_id=session_id, task_id=graph_key))
        self.write_rendered(
            template,
            session_id=session_id,
            session_name=session_name,
            task_id=graph_key,
            task_progress_script=task_progress_script,
        )

    @staticmethod
    def task_progress_app(doc, scheduler_ip=None):
        session_id = to_str(doc.session_context.request.arguments.get('session_id')[0])
        task_id = to_str(doc.session_context.request.arguments.get('task_id')[0])
        web_api = MarsWebAPI(scheduler_ip)

        states = list(OperandState.__members__.values())

        ops, stats, progress = web_api.get_task_detail(session_id, task_id)
        source = ColumnDataSource(data=stats)
        cols = list(stats)[1:]
        p = figure(y_range=ops, plot_height=500, plot_width=800, x_range=(0, 100),
                   title="Total Progress: %0.2f%%" % progress)
        renderers = p.hbar_stack(cols, y='ops', height=0.9, color=_base_palette[0:len(states)],
                                 source=source)

        legend_items = [('\0' + s, [r]) for s, r in zip(cols, renderers)]
        legend = Legend(items=legend_items, location='bottom_right',
                        orientation='horizontal', label_text_font_size='10px')

        p.add_layout(legend, 'below')

        p.ygrid.grid_line_color = None
        p.yaxis.major_tick_line_color = None
        p.yaxis.minor_tick_line_color = None
        p.axis.minor_tick_line_color = None
        p.outline_line_color = None

        doc.add_root(p)

        def _refresher():
            try:
                _, new_stats, new_progress = web_api.get_task_detail(session_id, task_id)
            except ActorNotExist:  # pragma: no cover
                doc.remove_periodic_callback(cb)
                p.title.text = "Graph not found, session may be stopped"
                return
            p.title.text = "Total Progress: %0.2f%%" % new_progress
            source.data = new_stats

            if new_progress >= 100.0:
                doc.remove_periodic_callback(cb)

        if progress < 100.0:
            cb = doc.add_periodic_callback(_refresher, 5000)


class TaskRunningNodesHandler(MarsRequestHandler):
    def get(self, session_id, graph_key):
        session_name = session_id

        running_info = self.web_api.get_operand_info(session_id, graph_key, OperandState.RUNNING)
        sorted_info = OrderedDict(
            sorted((item for item in running_info.items()), key=lambda it: (it[1].get('op_name'), it[0])))

        template = _jinja_env.get_template('task_pages/runnings.html')
        self.write_rendered(
            template,
            session_id=session_id,
            session_name=session_name,
            task_id=graph_key,
            running_info=sorted_info
        )


register_web_handler('/session', TaskListHandler)
register_web_handler('/session/(?P<session_id>[^/]+)/graph/(?P<graph_key>[^/]+)',
                     TaskHandler)
register_web_handler('/session/(?P<session_id>[^/]+)/graph/(?P<graph_key>[^/]+)/running_nodes',
                     TaskRunningNodesHandler)

register_bokeh_app('/' + PROGRESS_APP_NAME, TaskHandler.task_progress_app)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...core import Base, Entity
from ...custom_log import redirect_custom_log
from ...serialize import FunctionField, BoolField, ListField, \
    DictField, StringField
from ...tiles import TilesError
from ...utils import enter_current_session, quiet_stdio, \
    find_objects, replace_objects, check_chunks_unknown_shape
from ..operands import TensorOperand, TensorOperandMixin


class TensorMapChunk(TensorOperand, TensorOperandMixin):
    _op_type_ = opcodes.MAP_CHUNK

    _func = FunctionField('func')
    _elementwise = BoolField('elementwise')
    _args = ListField('args')
    _kwargs = DictField('kwargs')
    _with_chunk_index = BoolField('with_chunk_index')
    # for chunk
    _tileable_op_key = StringField('tileable_op_key')

    def __init__(self, func=None, args=None, kwargs=None, tileable_op_key=None,
                 elementwise=None, with_chunk_index=None, **kw):
        args = list(args) if args is not None else None
        super().__init__(_func=func, _args=args, _kwargs=kwargs, _elementwise=elementwise,
                         _with_chunk_index=with_chunk_index,
                         _tileable_op_key=tileable_op_key, **kw)

    @property
    def func(self):
        return self._func

    @property
    def elementwise(self):
        return self._elementwise

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def tileable_op_key(self):
        return self._tileable_op_key

    @property
    def with_chunk_index(self):
        return self._with_chunk_index

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        old_inputs = find_objects(self._args, (Base, Entity)) \
            + find_objects(self._kwargs, (Base, Entity))
        mapping = {o: n for o, n in zip(old_inputs, self._inputs[1:])}
        self._args = replace_objects(self._args, mapping)
        self._kwargs = replace_objects(self._kwargs, mapping)

    def __call__(self, t, dtype=None):
        if dtype is None:
            try:
                kwargs = self.kwargs or dict()
                if self.with_chunk_index:
                    kwargs['chunk_index'] = (0,) * t.ndim
                with np.errstate(all='ignore'), quiet_stdio():
                    mock_result = self.func(np.random.rand(2, 2).astype(t.dtype),
                                            *(self.args or ()), **kwargs)
            except:
                raise TypeError('Cannot estimate output type of map_chunk call')
            dtype = mock_result.dtype

        new_shape = t.shape if self.elementwise else (np.nan,) * t.ndim
        inputs = [t] + find_objects(self.args, (Base, Entity)) + \
            find_objects(self.kwargs, (Base, Entity))
        return self.new_tensor(inputs, dtype=dtype, shape=new_shape)

    @classmethod
    def tile(cls, op: 'TensorMapChunk'):
        inp = op.inputs[0]
        out = op.outputs[0]

        new_inputs = [op.inputs[0]]
        check_chunks_unknown_shape(op.inputs[1:], TilesError)
        for other_inp in op.inputs[1:]:
            new_inputs.append(other_inp.rechunk(other_inp.shape)._inplace_tile())

        chunks = []
        for c in inp.chunks:
            params = c.params
            params['dtype'] = inp.dtype
            if not op.elementwise:
                params['shape'] = (np.nan,) * c.ndim

            new_op = op.copy().reset_key()
            new_op._tileable_op_key = out.key
            chunk_inputs = [c]
            for other_inp in new_inputs[1:]:
                chunk_inputs.append(other_inp.chunks[0])
            chunks.append(new_op.new_chunk(chunk_inputs, **params))

        new_op = op.copy().reset_key()
        params = out.params
        nsplits = inp.nsplits
        if not op.elementwise:
            nsplits = tuple((np.nan,) * len(sp) for sp in nsplits)
        return new_op.new_tileables([inp], chunks=chunks, nsplits=nsplits, **params)

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op: 'TensorMapChunk'):
        in_data = ctx[op.inputs[0].key]
        out_chunk = op.outputs[0]

        args = op.args or tuple()
        kwargs = op.kwargs or dict()
        if op.with_chunk_index:
            kwargs['chunk_index'] = out_chunk.index

        chunks = find_objects(args, (Base, Entity)) + find_objects(kwargs, (Base, Entity))
        mapping = {chunk: ctx[chunk.key] for chunk in chunks}
        args = replace_objects(args, mapping)
        kwargs = replace_objects(kwargs, mapping)

        ctx[op.outputs[0].key] = op.func(in_data, *args, **kwargs)


def map_chunk(t, func, args=(), **kwargs):
    """
    Apply function to each chunk.

    Parameters
    ----------
    func : function
        Function to apply to each chunk.
    args : tuple
        Positional arguments to pass to func in addition to the array.
    **kwargs
        Additional keyword arguments to pass as keywords arguments to func.

    Returns
    -------
    Tensor
        Result of applying ``func`` to each chunk of the Tensor.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.array([[4, 9]] * 3)
    >>> a.execute()
    array([[4, 9],
           [4, 9],
           [4, 9]])

    Output dtype will be auto inferred.

    >>> a.map_chunk(lambda c: c * 0.5).execute()
    array([[2. , 4.5],
           [2. , 4.5],
           [2. , 4.5]])

    You can specify ``dtype`` by yourself if auto infer failed.
    """
    elementwise = kwargs.pop('elementwise', None)
    dtype = np.dtype(kwargs.pop('dtype')) if 'dtype' in kwargs else None
    with_chunk_index = kwargs.pop('with_chunk_index', False)

    op = TensorMapChunk(func=func, args=args, kwargs=kwargs, elementwise=elementwise,
                        with_chunk_index=with_chunk_index)
    return op(t, dtype=dtype)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import AnyField
from .core import TensorRandomOperandMixin, handle_array, TensorDistribution


class TensorNoncentralChisquare(TensorDistribution, TensorRandomOperandMixin):
    __slots__ = '_df', '_nonc', '_size'
    _input_fields_ = ['_df', '_nonc']
    _op_type_ = OperandDef.RAND_NONCENTRAL_CHISQURE

    _df = AnyField('df')
    _nonc = AnyField('nonc')
    _func_name = 'noncentral_chisquare'

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super().__init__(_size=size, _state=state, _dtype=dtype, _gpu=gpu, **kw)

    @property
    def df(self):
        return self._df

    @property
    def nonc(self):
        return self._nonc

    def __call__(self, df, nonc, chunk_size=None):
        return self.new_tensor([df, nonc], None, raw_chunk_size=chunk_size)


def noncentral_chisquare(random_state, df, nonc, size=None, chunk_size=None, gpu=None, dtype=None):
    r"""
    Draw samples from a noncentral chi-square distribution.

    The noncentral :math:`\chi^2` distribution is a generalisation of
    the :math:`\chi^2` distribution.

    Parameters
    ----------
    df : float or array_like of floats
        Degrees of freedom, should be > 0.
    nonc : float or array_like of floats
        Non-centrality, should be non-negative.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``df`` and ``nonc`` are both scalars.
        Otherwise, ``mt.broadcast(df, nonc).size`` samples are drawn.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized noncentral chi-square distribution.

    Notes
    -----
    The probability density function for the noncentral Chi-square
    distribution is

    .. math:: P(x;df,nonc) = \sum^{\infty}_{i=0}
                           \frac{e^{-nonc/2}(nonc/2)^{i}}{i!}
                           \P_{Y_{df+2i}}(x),

    where :math:`Y_{q}` is the Chi-square with q degrees of freedom.

    In Delhi (2007), it is noted that the noncentral chi-square is
    useful in bombing and coverage problems, the probability of
    killing the point target given by the noncentral chi-squared
    distribution.

    References
    ----------
    .. [1] Delhi, M.S. Holla, "On a noncentral chi-square distribution in
           the analysis of weapon systems effectiveness", Metrika,
           Volume 15, Number 1 / December, 1970.
    .. [2] Wikipedia, "Noncentral chi-square distribution"
           http://en.wikipedia.org/wiki/Noncentral_chi-square_distribution

    Examples
    --------
    Draw values from the distribution and plot the histogram

    >>> import matplotlib.pyplot as plt
    >>> import mars.tensor as mt
    >>> values = plt.hist(mt.random.noncentral_chisquare(3, 20, 100000).execute(),
    ...                   bins=200, normed=True)
    >>> plt.show()

    Draw values from a noncentral chisquare with very small noncentrality,
    and compare to a chisquare.

    >>> plt.figure()
    >>> values = plt.hist(mt.random.noncentral_chisquare(3, .0000001, 100000).execute(),
    ...                   bins=mt.arange(0., 25, .1).execute(), normed=True)
    >>> values2 = plt.hist(mt.random.chisquare(3, 100000).execute(),
    ...                    bins=mt.arange(0., 25, .1).execute(), normed=True)
    >>> plt.plot(values[1][0:-1], values[0]-values2[0], 'ob')
    >>> plt.show()

    Demonstrate how large values of non-centrality lead to a more symmetric
    distribution.

    >>> plt.figure()
    >>> values = plt.hist(mt.random.noncentral_chisquare(3, 20, 100000).execute(),
    ...                   bins=200, normed=True)
    >>> plt.show()
    """
    if dtype is None:
        dtype = np.random.RandomState().noncentral_chisquare(
            handle_array(df), handle_array(nonc), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorNoncentralChisquare(size=size, state=random_state.to_numpy(), gpu=gpu, dtype=dtype)
    return op(df, nonc, chunk_size=chunk_size)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable
from functools import partial

import numpy as np

from .. import opcodes
from ..core import Entity, Base, ChunkData
from ..custom_log import redirect_custom_log
from ..dataframe.core import DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE
from ..serialize import FunctionField, ListField, DictField, \
    BoolField, Int32Field, StringField
from ..operands import ObjectOperand
from ..tensor.core import TENSOR_TYPE
from ..utils import build_fetch_tileable, calc_nsplits, \
    enter_current_session, find_objects, replace_objects
from .operands import RemoteOperandMixin


class _TileablePlaceholder:
    def __init__(self, tileable):
        self.tileable = build_fetch_tileable(tileable)

    def __getstate__(self):
        fetch_op = self.tileable.op
        fetch_tileable = self.tileable
        chunk_infos = [(type(c.op), c.op.output_types, c.key, c.id, c.params)
                       for c in fetch_tileable.chunks]
        return type(fetch_op), fetch_op.id, fetch_op.output_types, \
               fetch_tileable.params, fetch_tileable.nsplits, chunk_infos

    def __setstate__(self, state):
        fetch_op_type, fetch_op_id, output_types, params, nsplits, chunk_infos = state
        params['nsplits'] = nsplits
        chunks = []
        for ci in chunk_infos:
            chunk_op_type, chunk_op_output_types, chunk_key, chunk_id, chunk_params = ci
            chunk = chunk_op_type(output_types=chunk_op_output_types).new_chunk(
                None, _key=chunk_key, _id=chunk_id, kws=[chunk_params])
            chunks.append(chunk)
        params['chunks'] = chunks
        self.tileable = fetch_op_type(
            _id=fetch_op_id, output_types=output_types).new_tileable(None, kws=[params])

    def __mars_tokenize__(self):
        return self.__getstate__()

    def __call__(self):  # pragma: no cover
        # make itself serializable
        pass


class RemoteFunction(RemoteOperandMixin, ObjectOperand):
    _op_type_ = opcodes.REMOTE_FUNCATION
    _op_module_ = 'remote'

    _function = FunctionField('function')
    _function_args = ListField('function_args')
    _function_kwargs = DictField('function_kwargs')
    _retry_when_fail = BoolField('retry_when_fail')
    _n_output = Int32Field('n_output')
    # for chunk
    _tileable_op_key = StringField('tileable_op_key')

    def __init__(self, function=None, function_args=None,
                 function_kwargs=None, retry_when_fail=None,
                 n_output=None, tileable_op_key=None, **kw):
        super().__init__(_function=function, _function_args=function_args,
                         _function_kwargs=function_kwargs,
                         _retry_when_fail=retry_when_fail,
                         _n_output=n_output, _tileable_op_key=tileable_op_key, **kw)

    @property
    def function(self):
        return self._function

    @property
    def function_args(self):
        return self._function_args

    @property
    def function_kwargs(self):
        return self._function_kwargs

    @property
    def retry_when_fail(self):
        return self._retry_when_fail

    @property
    def n_output(self):
        return self._n_output

    @property
    def tileable_op_key(self):
        return self._tileable_op_key

    @property
    def output_limit(self):
        return self._n_output or 1

    @property
    def retryable(self) -> bool:
        return self._retry_when_fail

    @classmethod
    def _no_prepare(cls, tileable):
        return isinstance(tileable, (TENSOR_TYPE, DATAFRAME_TYPE,
                                     SERIES_TYPE, INDEX_TYPE))

    def _set_inputs(self, inputs):
        raw_inputs = getattr(self, '_inputs', None)
        super()._set_inputs(inputs)

        function_inputs = iter(inp for inp in self._inputs)
        mapping = {inp: new_inp for inp, new_inp in zip(inputs, self._inputs)}
        if raw_inputs is not None:
            for raw_inp in raw_inputs:
                if self._no_prepare(raw_inp):
                    if not isinstance(self._inputs[0], ChunkData):
                        # not in tile, set_inputs from tileable
                        mapping[raw_inp] = next(function_inputs)
                    else:
                        # in tile, set_inputs from chunk
                        mapping[raw_inp] = _TileablePlaceholder(raw_inp)
                else:
                    mapping[raw_inp] = next(function_inputs)
        self._function_args = replace_objects(self._function_args, mapping)
        self._function_kwargs = replace_objects(self._function_kwargs, mapping)

    def __call__(self):
        find_inputs = partial(find_objects, types=(Entity, Base))
        inputs = find_inputs(self._function_args) + find_inputs(self._function_kwargs)
        if self.n_output is None:
            return self.new_tileable(inputs)
        else:
            return self.new_tileables(
                inputs, kws=[dict(i=i) for i in range(self.n_output)])

    @classmethod
    def tile(cls, op):
        outs = op.outputs
        chunk_op = op.copy().reset_key()

        chunk_inputs = []
        pure_depends = []
        for inp in op.inputs:
            if cls._no_prepare(inp):  # pragma: no cover
                # if input is tensor, DataFrame etc,
                # do not prepare data, because the data may be to huge,
                # and users can choose to fetch slice of the data themselves
                pure_depends.extend([True] * len(inp.chunks))
            else:
                pure_depends.extend([False] * len(inp.chunks))
            chunk_inputs.extend(inp.chunks)
        chunk_op._pure_depends = pure_depends
        # record tileable op key for chunk op
        chunk_op._tileable_op_key = op.key

        out_chunks = [list() for _ in range(len(outs))]
        chunk_kws = []
        for i, out in enumerate(outs):
            chunk_params = out.params
            chunk_params['index'] = ()
            chunk_params['i'] = i
            chunk_kws.append(chunk_params)
        chunks = chunk_op.new_chunks(chunk_inputs, kws=chunk_kws)
        for i, c in enumerate(chunks):
            out_chunks[i].append(c)

        kws = []
        for i, out in enumerate(outs):
            params = out.params
            params['chunks'] = out_chunks[i]
            params['nsplits'] = ()
            kws.append(params)
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=kws)

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op: "RemoteFunction"):
        mapping = {inp: ctx[inp.key] for inp, is_pure_dep
                   in zip(op.inputs, op.pure_depends) if not is_pure_dep}
        for to_search in [op.function_args, op.function_kwargs]:
            tileable_placeholders = find_objects(to_search, _TileablePlaceholder)
            for ph in tileable_placeholders:
                tileable = ph.tileable
                chunk_index_to_shape = dict()
                for chunk in tileable.chunks:
                    if any(np.isnan(s) for s in chunk.shape):
                        shape = ctx.get_chunk_metas([chunk.key], filter_fields=['chunk_shape'])[0][0]
                        chunk._shape = shape
                    chunk_index_to_shape[chunk.index] = chunk.shape
                if any(any(np.isnan(s) for s in ns) for ns in tileable._nsplits):
                    nsplits = calc_nsplits(chunk_index_to_shape)
                    tileable._nsplits = nsplits
                    tileable._shape = tuple(sum(ns) for ns in nsplits)
                mapping[ph] = tileable

        function = op.function
        function_args = replace_objects(op.function_args, mapping)
        function_kwargs = replace_objects(op.function_kwargs, mapping)

        result = function(*function_args, **function_kwargs)

        if op.n_output is None:
            ctx[op.outputs[0].key] = result
        else:
            if not isinstance(result, Iterable):
                raise TypeError(f'Specifying n_output={op.n_output}, '
                                f'but result is not iterable, got {result}')
            result = list(result)
            if len(result) != op.n_output:
                raise ValueError(f'Length of return value should be {op.n_output}, '
                                 f'got {len(result)}')
            for out, r in zip(op.outputs, result):
                ctx[out.key] = r


def spawn(func, args=(), kwargs=None, retry_when_fail=False, n_output=None):
    """
    Spawn a function and return a Mars Object which can be executed later.

    Parameters
    ----------
    func : function
        Function to spawn.
    args: tuple
       Args to pass to function
    kwargs: dict
       Kwargs to pass to function
    retry_when_fail: bool, default False
       If True, retry when function failed.
    n_output: int
       Count of outputs for the function

    Returns
    -------
    Object
        Mars Object.

    Examples
    --------
    >>> import mars.remote as mr
    >>> def inc(x):
    >>>     return x + 1
    >>>
    >>> result = mr.spawn(inc, args=(0,))
    >>> result
    Object <op=RemoteFunction, key=e0b31261d70dd9b1e00da469666d72d9>
    >>> result.execute().fetch()
    1

    List of spawned functions can be converted to :class:`mars.remote.ExecutableTuple`,
    and `.execute()` can be called to run together.

    >>> results = [mr.spawn(inc, args=(i,)) for i in range(10)]
    >>> mr.ExecutableTuple(results).execute().fetch()
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    Mars Object returned by :meth:`mars.remote.spawn` can be treated
    as arguments for other spawn functions.

    >>> results = [mr.spawn(inc, args=(i,)) for i in range(10)]   # list of spawned functions
    >>> def sum_all(xs):
            return sum(xs)
    >>> mr.spawn(sum_all, args=(results,)).execute().fetch()
    55

    inside a spawned function, new functions can be spawned.

    >>> def driver():
    >>>     results = [mr.spawn(inc, args=(i,)) for i in range(10)]
    >>>     return mr.ExecutableTuple(results).execute().fetch()
    >>>
    >>> mr.spawn(driver).execute().fetch()
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    Mars tensor, DataFrame and so forth is available in spawned functions as well.

    >>> import mars.tensor as mt
    >>> def driver2():
    >>>     t = mt.random.rand(10, 10)
    >>>     return t.sum().to_numpy()
    >>>
    >>> mr.spawn(driver2).execute().fetch()
    52.47844223908132

    Argument of `n_output` can indicate that the spawned function will return multiple outputs.
    This is important when some of the outputs may be passed to different functions.

    >>> def triage(alist):
    >>>     ret = [], []
    >>>     for i in alist:
    >>>         if i < 0.5:
    >>>             ret[0].append(i)
    >>>         else:
    >>>             ret[1].append(i)
    >>>     return ret
    >>>
    >>> def sum_all(xs):
    >>>     return sum(xs)
    >>>
    >>> l = [0.4, 0.7, 0.2, 0.8]
    >>> la, lb = mr.spawn(triage, args=(l,), n_output=2)
    >>>
    >>> sa = mr.spawn(sum_all, args=(la,))
    >>> sb = mr.spawn(sum_all, args=(lb,))
    >>> mr.ExecutableTuple([sa, sb]).execute().fetch()
    >>> [0.6000000000000001, 1.5]
    """
    if not isinstance(args, tuple):
        args = [args]
    else:
        args = list(args)
    if kwargs is None:
        kwargs = dict()
    if not isinstance(kwargs, dict):
        raise TypeError('kwargs has to be a dict')

    op = RemoteFunction(function=func, function_args=args,
                        function_kwargs=kwargs,
                        retry_when_fail=retry_when_fail,
                        n_output=n_output)
    return op()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import cloudpickle

from ... import opcodes
from ...serialize import StringField, AnyField, BoolField, \
    Int64Field, BytesField
from ..core import DATAFRAME_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index, build_empty_df, build_empty_series, create_sa_connection


class DataFrameToSQLTable(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.TO_SQL

    _table_name = StringField('table_name')
    _con = StringField('con')
    _schema = StringField('schema')
    _if_exists = StringField('if_exists')
    _index = BoolField('index')
    _index_label = AnyField('index_label')
    _chunksize = Int64Field('chunksize')
    _dtype = AnyField('dtype')
    _method = AnyField('method')
    _engine_kwargs = BytesField('engine_kwargs', on_serialize=cloudpickle.dumps,
                                on_deserialize=cloudpickle.loads)

    def __init__(self, table_name=None, con=None, schema=None, if_exists=None, index=None,
                 index_label=None, chunksize=None, dtype=None, method=None, engine_kwargs=None, **kw):
        super().__init__(_table_name=table_name, _con=con, _schema=schema, _if_exists=if_exists,
                         _index=index, _index_label=index_label, _chunksize=chunksize,
                         _dtype=dtype, _method=method, _engine_kwargs=engine_kwargs, **kw)

    @property
    def table_name(self):
        return self._table_name

    @property
    def con(self):
        return self._con

    @property
    def schema(self):
        return self._schema

    @property
    def if_exists(self):
        return self._if_exists

    @property
    def index(self):
        return self._index

    @property
    def index_label(self):
        return self._index_label

    @property
    def chunksize(self):
        return self._chunksize

    @property
    def dtype(self):
        return self._dtype

    @property
    def method(self):
        return self._method

    @property
    def engine_kwargs(self):
        return self._engine_kwargs

    def __call__(self, df_or_series):
        with create_sa_connection(self._con, **(self._engine_kwargs or dict())) as con:
            self._con = str(con.engine.url)
            empty_index = df_or_series.index_value.to_pandas()[:0]
            if isinstance(df_or_series, DATAFRAME_TYPE):
                empty_obj = build_empty_df(df_or_series.dtypes, index=empty_index)
            else:
                empty_obj = build_empty_series(df_or_series.dtype, index=empty_index, name=df_or_series.name)

            empty_obj.to_sql(self.table_name, con=con, schema=self.schema, if_exists=self.if_exists,
                             index=self.index, index_label=self.index_label, dtype=self.dtype)

            index_value = parse_index(df_or_series.index_value.to_pandas()[:0], df_or_series.key, 'index')
            if isinstance(df_or_series, DATAFRAME_TYPE):
                columns_value = parse_index(df_or_series.columns_value.to_pandas()[:0],
                                            df_or_series.key, 'columns', store_data=True)
                return self.new_dataframe([df_or_series], shape=(0, 0), dtypes=df_or_series.dtypes[:0],
                                          index_value=index_value, columns_value=columns_value)
            else:
                return self.new_series([df_or_series], shape=(0,), dtype=df_or_series.dtype,
                                       index_value=index_value)

    @classmethod
    def tile(cls, op: 'DataFrameToSQLTable'):
        inp = op.inputs[0]
        out = op.outputs[0]
        if inp.ndim == 2:
            inp = inp.rechunk({1: (inp.shape[1],)})._inplace_tile()

        chunks = []
        for c in inp.chunks:
            new_op = op.copy().reset_key()
            new_op._if_exists = 'append'

            index_value = parse_index(c.index_value.to_pandas()[:0], c)
            if c.ndim == 2:
                columns_value = parse_index(c.columns_value.to_pandas()[:0], store_data=True)
                chunks.append(new_op.new_chunk([c], shape=(0, 0), index=c.index, dtypes=out.dtypes,
                                               index_value=index_value, columns_value=columns_value))
            else:
                chunks.append(new_op.new_chunk([c], shape=(0,), index=c.index, dtype=out.dtype,
                                               index_value=index_value))

        new_op = op.copy().reset_key()
        params = out.params.copy()
        params['nsplits'] = tuple((0,) * len(sp) for sp in inp.nsplits)
        return new_op.new_tileables([inp], chunks=chunks, **params)

    @classmethod
    def execute(cls, ctx, op: 'DataFrameToSQLTable'):
        in_df = op.inputs[0]
        out_df = op.outputs[0]
        in_data = ctx[in_df.key]

        import sqlalchemy as sa
        engine = sa.create_engine(op.con, **(op.engine_kwargs or dict()))

        try:
            with engine.connect() as connection:
                with connection.begin():
                    in_data.to_sql(op.table_name, con=connection, if_exists=op.if_exists, index=op.index,
                                   index_label=op.index_label, chunksize=op.chunksize, dtype=op.dtype,
                                   method=op.method)
        finally:
            engine.dispose()

        if in_df.ndim == 2:
            ctx[out_df.key] = pd.DataFrame()
        else:
            ctx[out_df.key] = pd.Series([], dtype=in_data.dtype)


def to_sql(df, name: str, con, schema=None, if_exists: str = 'fail', index: bool = True, index_label=None,
           chunksize=None, dtype=None, method=None):
    """
    Write records stored in a DataFrame to a SQL database.

    Databases supported by SQLAlchemy [1]_ are supported. Tables can be
    newly created, appended to, or overwritten.

    Parameters
    ----------
    name : str
        Name of SQL table.
    con : sqlalchemy.engine.Engine or sqlite3.Connection
        Using SQLAlchemy makes it possible to use any DB supported by that
        library. Legacy support is provided for sqlite3.Connection objects. The user
        is responsible for engine disposal and connection closure for the SQLAlchemy
        connectable See `here                 <https://docs.sqlalchemy.org/en/13/core/connections.html>`_

    schema : str, optional
        Specify the schema (if database flavor supports this). If None, use
        default schema.
    if_exists : {'fail', 'replace', 'append'}, default 'fail'
        How to behave if the table already exists.

        * fail: Raise a ValueError.
        * replace: Drop the table before inserting new values.
        * append: Insert new values to the existing table.

    index : bool, default True
        Write DataFrame index as a column. Uses `index_label` as the column
        name in the table.
    index_label : str or sequence, default None
        Column label for index column(s). If None is given (default) and
        `index` is True, then the index names are used.
        A sequence should be given if the DataFrame uses MultiIndex.
    chunksize : int, optional
        Specify the number of rows in each batch to be written at a time.
        By default, all rows will be written at once.
    dtype : dict or scalar, optional
        Specifying the datatype for columns. If a dictionary is used, the
        keys should be the column names and the values should be the
        SQLAlchemy types or strings for the sqlite3 legacy mode. If a
        scalar is provided, it will be applied to all columns.
    method : {None, 'multi', callable}, optional
        Controls the SQL insertion clause used:

        * None : Uses standard SQL ``INSERT`` clause (one per row).
        * 'multi': Pass multiple values in a single ``INSERT`` clause.
        * callable with signature ``(pd_table, conn, keys, data_iter)``.

        Details and a sample callable implementation can be found in the
        section :ref:`insert method <io.sql.method>`.

        .. versionadded:: 0.24.0

    Raises
    ------
    ValueError
        When the table already exists and `if_exists` is 'fail' (the
        default).

    See Also
    --------
    read_sql : Read a DataFrame from a table.

    Notes
    -----
    Timezone aware datetime columns will be written as
    ``Timestamp with timezone`` type with SQLAlchemy if supported by the
    database. Otherwise, the datetimes will be stored as timezone unaware
    timestamps local to the original timezone.

    .. versionadded:: 0.24.0

    References
    ----------
    .. [1] http://docs.sqlalchemy.org
    .. [2] https://www.python.org/dev/peps/pep-0249/

    Examples
    --------

    Create an in-memory SQLite database.

    >>> import mars.dataframe as md
    >>> from sqlalchemy import create_engine
    >>> engine = create_engine('sqlite:////tmp/temp.db')

    Create a table from scratch with 3 rows.

    >>> df = md.DataFrame({'name' : ['User 1', 'User 2', 'User 3']})
    >>> df.execute()
         name
    0  User 1
    1  User 2
    2  User 3

    >>> df.to_sql('users', con=engine).execute()
    >>> engine.execute("SELECT * FROM users").fetchall()
    [(0, 'User 1'), (1, 'User 2'), (2, 'User 3')]

    >>> df1 = md.DataFrame({'name' : ['User 4', 'User 5']})
    >>> df1.to_sql('users', con=engine, if_exists='append').execute()
    >>> engine.execute("SELECT * FROM users").fetchall()
    [(0, 'User 1'), (1, 'User 2'), (2, 'User 3'),
     (0, 'User 4'), (1, 'User 5')]

    Overwrite the table with just ``df1``.

    >>> df1.to_sql('users', con=engine, if_exists='replace',
    ...            index_label='id').execute()
    >>> engine.execute("SELECT * FROM users").fetchall()
    [(0, 'User 4'), (1, 'User 5')]

    Specify the dtype (especially useful for integers with missing values).
    Notice that while pandas is forced to store the data as floating point,
    the database supports nullable integers. When fetching the data with
    Python, we get back integer scalars.

    >>> df = md.DataFrame({"A": [1, None, 2]})
    >>> df.execute()
         A
    0  1.0
    1  NaN
    2  2.0

    >>> from sqlalchemy.types import Integer
    >>> df.to_sql('integers', con=engine, index=False,
    ...           dtype={"A": Integer()}).execute()

    >>> engine.execute("SELECT * FROM integers").fetchall()
    [(1,), (None,), (2,)]
    """
    op = DataFrameToSQLTable(table_name=name, con=con, schema=schema, if_exists=if_exists, index=index,
                             index_label=index_label, chunksize=chunksize, dtype=dtype, method=method)
    return op(df)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...config import options
from .core import DataFrameCumReductionOperand, DataFrameCumReductionMixin


class DataFrameCummin(DataFrameCumReductionOperand, DataFrameCumReductionMixin):
    _op_type_ = OperandDef.CUMMIN
    _func_name = 'cummin'


def cummin(df, axis=None, skipna=True):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameCummin(axis=axis, skipna=skipna, output_types=df.op.output_types,
                         use_inf_as_na=use_inf_as_na)
    return op(df)
